from typing import TYPE_CHECKING

import claripy
import pyvex

from ...engines.vex.claripy.datalayer import value as claripy_value
from ...engines.light import SimEngineLightVEXMixin
from ..typehoon import typevars, typeconsts
from .engine_base import SimEngineVRBase, RichR

if TYPE_CHECKING:
    from .variable_recovery_base import VariableRecoveryStateBase


class SimEngineVRVEX(
    SimEngineLightVEXMixin,
    SimEngineVRBase,
):
    state: 'VariableRecoveryStateBase'

    # Statement handlers

    def _handle_Put(self, stmt):
        offset = stmt.offset
        r = self._expr(stmt.data)
        size = stmt.data.result_size(self.tyenv) // 8

        if offset == self.arch.ip_offset:
            return
        self._assign_to_register(offset, r, size)

    def _handle_Store(self, stmt):
        addr_r = self._expr(stmt.addr)
        size = stmt.data.result_size(self.tyenv) // 8
        r = self._expr(stmt.data)

        self._store(addr_r, r, size, stmt=stmt)

    def _handle_StoreG(self, stmt):
        guard = self._expr(stmt.guard)
        if guard is True:

            addr = self._expr(stmt.addr)
            size = stmt.data.result_size(self.tyenv) // 8
            data = self._expr(stmt.data)
            self._store(addr, data, size, stmt=stmt)

    def _handle_LoadG(self, stmt):
        guard = self._expr(stmt.guard)
        if guard is True:
            addr = self._expr(stmt.addr)
            if addr is not None:
                self.tmps[stmt.dst] = self._load(addr, self.tyenv.sizeof(stmt.dst) // 8)
        elif guard is False:
            data = self._expr(stmt.alt)
            self.tmps[stmt.dst] = data
        else:
            self.tmps[stmt.dst] = None

    def _handle_LLSC(self, stmt: pyvex.IRStmt.LLSC):
        if stmt.storedata is None:
            # load-link
            addr = self._expr(stmt.addr)
            size = self.tyenv.sizeof(stmt.result) // self.arch.byte_width
            data = self._load(addr, size)
            self.tmps[stmt.result] = data
        else:
            # store-conditional
            storedata = self._expr(stmt.storedata)
            addr = self._expr(stmt.addr)
            size = self.tyenv.sizeof(stmt.storedata.tmp) // self.arch.byte_width

            self._store(addr, storedata, size)

            result_size = self.tyenv.sizeof(stmt.result)
            self.tmps[stmt.result] = RichR(claripy.BVV(1, result_size))

    def _handle_NoOp(self, stmt):
        pass

    # Expression handlers

    def _expr(self, expr) -> RichR:
        """

        :param expr:
        :return:
        :rtype: RichR
        """

        r = super()._expr(expr)
        if r is None:
            bits = expr.result_size(self.tyenv)
            return RichR(self.state.top(bits))
        return r

    def _handle_Get(self, expr):
        reg_offset = expr.offset
        reg_size = expr.result_size(self.tyenv) // 8

        return self._read_from_register(reg_offset, reg_size, expr=expr)

    def _handle_Load(self, expr: pyvex.IRExpr.Load) -> RichR:
        addr = self._expr(expr.addr)
        size = expr.result_size(self.tyenv) // 8

        return self._load(addr, size)

    def _handle_CCall(self, expr):  # pylint:disable=useless-return
        # ccalls don't matter
        return RichR(self.state.top(expr.result_size(self.tyenv)))

    def _handle_Conversion(self, expr: pyvex.IRExpr.Unop) -> RichR:
        return RichR(self.state.top(expr.result_size(self.tyenv)))


    # Function handlers

    def _handle_function(self, func_addr):  # pylint:disable=unused-argument,no-self-use,useless-return
        return None

    def _handle_Const(self, expr):
        return RichR(claripy_value(expr.con.type, expr.con.value), typevar=typeconsts.int_type(expr.con.size))

    def _handle_Add(self, expr):
        arg0, arg1 = expr.args
        r0 = self._expr(arg0)
        r1 = self._expr(arg1)

        result_size = expr.result_size(self.tyenv)
        if r0.data.concrete and r1.data.concrete:
            # constants
            return RichR(r0.data + r1.data,
                         typevar=typeconsts.int_type(result_size),
                         type_constraints=None)

        typevar = None
        if r0.typevar is not None and r1.data.concrete:
            typevar = typevars.DerivedTypeVariable(r0.typevar, typevars.AddN(r1.data._model_concrete.value))

        sum_ = r0.data + r1.data
        return RichR(sum_,
                     typevar=typevar,
                     type_constraints={ typevars.Subtype(r0.typevar, r1.typevar) },
                     )

    def _handle_Sub(self, expr):
        arg0, arg1 = expr.args
        r0 = self._expr(arg0)
        r1 = self._expr(arg1)

        result_size = expr.result_size(self.tyenv)
        if r0.data.concrete and r1.data.concrete:
            # constants
            return RichR(r0.data - r1.data,
                         typevar=typeconsts.int_type(result_size),
                         type_constraints=None)

        typevar = None
        if r0.typevar is not None and r1.data.concrete:
            typevar = typevars.DerivedTypeVariable(r0.typevar, typevars.SubN(r1.data._model_concrete.value))

        diff = r0.data - r1.data
        return RichR(diff,
                     typevar=typevar,
                     )

    def _handle_And(self, expr):
        arg0, arg1 = expr.args
        r0 = self._expr(arg0)
        r1 = self._expr(arg1)

        result_size = expr.result_size(self.tyenv)
        if r0.data.concrete and r1.data.concrete:
            # constants
            return RichR(r0.data & r1.data)

        if self.state.is_stack_address(r0.data):
            r = r0.data
        elif self.state.is_stack_address(r1.data):
            r = r1.data
        else:
            r = self.state.top(result_size)
        return RichR(r)

    def _handle_Xor(self, expr):
        arg0, arg1 = expr.args
        r0 = self._expr(arg0)
        r1 = self._expr(arg1)

        result_size = expr.result_size(self.tyenv)
        if r0.data.concrete and r1.data.concrete:
            # constants
            return RichR(r0.data ^ r1.data)

        r = self.state.top(result_size)
        return RichR(r)

    def _handle_Or(self, expr):
        arg0, arg1 = expr.args
        r0 = self._expr(arg0)
        r1 = self._expr(arg1)

        result_size = expr.result_size(self.tyenv)
        if r0.data.concrete and r1.data.concrete:
            # constants
            return RichR(r0.data | r1.data)

        r = self.state.top(result_size)
        return RichR(r)

    def _handle_Not(self, expr):
        arg = expr.args[0]
        r0 = self._expr(arg)

        result_size = expr.result_size(self.tyenv)
        if r0.data.concrete:
            # constants
            return RichR(~r0.data)

        r = self.state.top(result_size)
        return RichR(r)

    def _handle_Mul(self, expr):
        arg0, arg1 = expr.args
        r0 = self._expr(arg0)
        r1 = self._expr(arg1)

        result_size = expr.result_size(self.tyenv)
        if r0.data.concrete and r1.data.concrete:
            # constants
            return RichR(r0.data * r1.data)

        r = self.state.top(result_size)
        return RichR(r)

    def _handle_DivMod(self, expr):
        arg0, arg1 = expr.args
        r0 = self._expr(arg0)
        r1 = self._expr(arg1)

        result_size = expr.result_size(self.tyenv)
        if r0.data.concrete and r1.data.concrete:
            # constants
            try:
                signed = "U" in expr.op  # Iop_DivModU64to32 vs Iop_DivMod
                from_size = r0.data.size()
                to_size = r1.data.size()
                if signed:
                    quotient = (r0.data.SDiv(claripy.SignExt(from_size - to_size, r1.data)))
                    remainder = (r0.data.SMod(claripy.SignExt(from_size - to_size, r1.data)))
                    quotient_size = to_size
                    remainder_size = to_size
                    result = claripy.Concat(
                        claripy.Extract(remainder_size - 1, 0, remainder),
                        claripy.Extract(quotient_size - 1, 0, quotient)
                    )
                else:
                    quotient = (r0.data // claripy.ZeroExt(from_size - to_size, r1.data))
                    remainder = (r0.data % claripy.ZeroExt(from_size - to_size, r1.data))
                    quotient_size = to_size
                    remainder_size = to_size
                    result = claripy.Concat(
                        claripy.Extract(remainder_size - 1, 0, remainder),
                        claripy.Extract(quotient_size - 1, 0, quotient)
                    )

                return RichR(result)
            except ZeroDivisionError:
                pass

        r = self.state.top(result_size)
        return RichR(r)

    def _handle_Div(self, expr):
        arg0, arg1 = expr.args
        r0 = self._expr(arg0)
        r1 = self._expr(arg1)

        result_size = expr.result_size(self.tyenv)
        if r0.data.concrete and r1.data.concrete:
            # constants
            try:
                return RichR(r0.data / r1.data)
            except ZeroDivisionError:
                pass

        r = self.state.top(result_size)
        return RichR(r)

    def _handle_Shr(self, expr):
        arg0, arg1 = expr.args
        r0 = self._expr(arg0)
        r1 = self._expr(arg1)

        result_size = expr.result_size(self.tyenv)
        if r0.data.concrete and r1.data.concrete:
            # constants
            return RichR(claripy.LShR(r0.data, r1.data._model_concrete.value),
                         typevar=typeconsts.int_type(result_size),
                         type_constraints=None)

        r = self.state.top(result_size)
        return RichR(r,
                     typevar=r0.typevar,
                     )

    def _handle_Sar(self, expr):
        arg0, arg1 = expr.args
        r0 = self._expr(arg0)
        r1 = self._expr(arg1)

        result_size = expr.result_size(self.tyenv)
        if r0.data.concrete and r1.data.concrete:
            # constants
            return RichR(r0.data >> r1.data._model_concrete.value,
                         typevar=typeconsts.int_type(result_size),
                         type_constraints=None)

        r = self.state.top(result_size)
        return RichR(r,
                     typevar=r0.typevar,
                     )

    def _handle_Shl(self, expr):
        arg0, arg1 = expr.args
        r0 = self._expr(arg0)
        r1 = self._expr(arg1)

        result_size = expr.result_size(self.tyenv)
        if r0.data.concrete and r1.data.concrete:
            # constants
            return RichR(r0.data << r1.data._model_concrete.value,
                         typevar=typeconsts.int_type(result_size),
                         type_constraints=None)

        r = self.state.top(result_size)
        return RichR(r,
                     typevar=r0.typevar,
                     )

    def _handle_CmpEQ(self, expr):
        arg0, arg1 = expr.args
        _ = self._expr(arg0)
        _ = self._expr(arg1)

        return RichR(self.state.top(1))

    def _handle_CmpNE(self, expr):
        arg0, arg1 = expr.args
        _ = self._expr(arg0)
        _ = self._expr(arg1)

        return RichR(self.state.top(1))

    def _handle_CmpLE(self, expr):
        arg0, arg1 = expr.args
        _ = self._expr(arg0)
        _ = self._expr(arg1)

        return RichR(self.state.top(1))

    def _handle_CmpLT(self, expr):
        arg0, arg1 = expr.args
        _ = self._expr(arg0)
        _ = self._expr(arg1)

        return RichR(self.state.top(1))

    def _handle_CmpGE(self, expr):
        arg0, arg1 = expr.args
        _ = self._expr(arg0)
        _ = self._expr(arg1)

        return RichR(self.state.top(1))

    def _handle_CmpGT(self, expr):
        arg0, arg1 = expr.args
        _ = self._expr(arg0)
        _ = self._expr(arg1)

        return RichR(self.state.top(1))



###
### This file was automatically generated
###

from archinfo.arch import register_arch, Endness, Register

from .common import ArchPcode


class ArchPcode_Toy_BE_64_harvard(ArchPcode):
    name = 'Toy:BE:64:harvard'
    pcode_arch = 'Toy:BE:64:harvard'
    description = 'Toy (test) processor 64-bit big-endian Harvard'
    bits = 64
    ip_offset = 0x1078
    sp_offset = 0x1068
    bp_offset = sp_offset
    instruction_endness = Endness.BE
    register_list = [
        Register('r0', 8, 0x1000),
        Register('r0h', 4, 0x1000),
        Register('r0l', 4, 0x1004),
        Register('r1', 8, 0x1008),
        Register('r1h', 4, 0x1008),
        Register('r1l', 4, 0x100c),
        Register('r2', 8, 0x1010),
        Register('r2h', 4, 0x1010),
        Register('r2l', 4, 0x1014),
        Register('r3', 8, 0x1018),
        Register('r3h', 4, 0x1018),
        Register('r3l', 4, 0x101c),
        Register('r4', 8, 0x1020),
        Register('r4h', 4, 0x1020),
        Register('r4l', 4, 0x1024),
        Register('r5', 8, 0x1028),
        Register('r5h', 4, 0x1028),
        Register('r5l', 4, 0x102c),
        Register('r6', 8, 0x1030),
        Register('r6h', 4, 0x1030),
        Register('r6l', 4, 0x1034),
        Register('r7', 8, 0x1038),
        Register('r7h', 4, 0x1038),
        Register('r7l', 4, 0x103c),
        Register('r8', 8, 0x1040),
        Register('r8h', 4, 0x1040),
        Register('r8l', 4, 0x1044),
        Register('r9', 8, 0x1048),
        Register('r9h', 4, 0x1048),
        Register('r9l', 4, 0x104c),
        Register('r10', 8, 0x1050),
        Register('r10h', 4, 0x1050),
        Register('r10l', 4, 0x1054),
        Register('r11', 8, 0x1058),
        Register('r11h', 4, 0x1058),
        Register('r11l', 4, 0x105c),
        Register('r12', 8, 0x1060),
        Register('r12h', 4, 0x1060),
        Register('r12l', 4, 0x1064),
        Register('sp', 8, 0x1068),
        Register('sph', 4, 0x1068),
        Register('spl', 4, 0x106c),
        Register('lr', 8, 0x1070),
        Register('lrh', 4, 0x1070),
        Register('lrl', 4, 0x1074),
        Register('pc', 8, 0x1078, alias_names=('ip',)),
        Register('pch', 4, 0x1078),
        Register('pcl', 4, 0x107c),
        Register('c', 1, 0x1100),
        Register('z', 1, 0x1101),
        Register('n', 1, 0x1102),
        Register('v', 1, 0x1103)
    ]

register_arch(['toy:be:64:harvard'], 64, Endness.BE, ArchPcode_Toy_BE_64_harvard)



# pylint:disable=line-too-long
import logging

from ...sim_type import SimTypeFunction,     SimTypeShort, SimTypeInt, SimTypeLong, SimTypeLongLong, SimTypeDouble, SimTypeFloat,     SimTypePointer,     SimTypeChar,     SimStruct,     SimTypeFixedSizeArray,     SimTypeBottom,     SimUnion,     SimTypeBool
from ...calling_conventions import SimCCStdcall, SimCCMicrosoftAMD64
from .. import SIM_PROCEDURES as P
from . import SimLibrary


_l = logging.getLogger(name=__name__)


lib = SimLibrary()
lib.set_default_cc('X86', SimCCStdcall)
lib.set_default_cc('AMD64', SimCCMicrosoftAMD64)
lib.set_library_names("winhttp.dll")
prototypes = \
    {
        # 
        'WinHttpSetStatusCallback': SimTypeFunction([SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypePointer(SimTypeFunction([SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt"), label="UIntPtr", offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypeInt(signed=False, label="UInt32")], SimTypeBottom(label="Void"), arg_names=["hInternet", "dwContext", "dwInternetStatus", "lpvStatusInformation", "dwStatusInformationLength"]), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeInt(signed=False, label="UInt"), label="UIntPtr", offset=0)], SimTypePointer(SimTypeFunction([SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt"), label="UIntPtr", offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypeInt(signed=False, label="UInt32")], SimTypeBottom(label="Void"), arg_names=["hInternet", "dwContext", "dwInternetStatus", "lpvStatusInformation", "dwStatusInformationLength"]), offset=0), arg_names=["hInternet", "lpfnInternetCallback", "dwNotificationFlags", "dwReserved"]),
        # 
        'WinHttpTimeFromSystemTime': SimTypeFunction([SimTypePointer(SimStruct({"wYear": SimTypeShort(signed=False, label="UInt16"), "wMonth": SimTypeShort(signed=False, label="UInt16"), "wDayOfWeek": SimTypeShort(signed=False, label="UInt16"), "wDay": SimTypeShort(signed=False, label="UInt16"), "wHour": SimTypeShort(signed=False, label="UInt16"), "wMinute": SimTypeShort(signed=False, label="UInt16"), "wSecond": SimTypeShort(signed=False, label="UInt16"), "wMilliseconds": SimTypeShort(signed=False, label="UInt16")}, name="SYSTEMTIME", pack=False, align=None), offset=0), SimTypePointer(SimTypeChar(label="Char"), label="LPArray", offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["pst", "pwszTime"]),
        # 
        'WinHttpTimeToSystemTime': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimStruct({"wYear": SimTypeShort(signed=False, label="UInt16"), "wMonth": SimTypeShort(signed=False, label="UInt16"), "wDayOfWeek": SimTypeShort(signed=False, label="UInt16"), "wDay": SimTypeShort(signed=False, label="UInt16"), "wHour": SimTypeShort(signed=False, label="UInt16"), "wMinute": SimTypeShort(signed=False, label="UInt16"), "wSecond": SimTypeShort(signed=False, label="UInt16"), "wMilliseconds": SimTypeShort(signed=False, label="UInt16")}, name="SYSTEMTIME", pack=False, align=None), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["pwszTime", "pst"]),
        # 
        'WinHttpCrackUrl': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), label="LPArray", offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimStruct({"dwStructSize": SimTypeInt(signed=False, label="UInt32"), "lpszScheme": SimTypePointer(SimTypeChar(label="Char"), offset=0), "dwSchemeLength": SimTypeInt(signed=False, label="UInt32"), "nScheme": SimTypeInt(signed=False, label="WINHTTP_INTERNET_SCHEME"), "lpszHostName": SimTypePointer(SimTypeChar(label="Char"), offset=0), "dwHostNameLength": SimTypeInt(signed=False, label="UInt32"), "nPort": SimTypeShort(signed=False, label="UInt16"), "lpszUserName": SimTypePointer(SimTypeChar(label="Char"), offset=0), "dwUserNameLength": SimTypeInt(signed=False, label="UInt32"), "lpszPassword": SimTypePointer(SimTypeChar(label="Char"), offset=0), "dwPasswordLength": SimTypeInt(signed=False, label="UInt32"), "lpszUrlPath": SimTypePointer(SimTypeChar(label="Char"), offset=0), "dwUrlPathLength": SimTypeInt(signed=False, label="UInt32"), "lpszExtraInfo": SimTypePointer(SimTypeChar(label="Char"), offset=0), "dwExtraInfoLength": SimTypeInt(signed=False, label="UInt32")}, name="URL_COMPONENTS", pack=False, align=None), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["pwszUrl", "dwUrlLength", "dwFlags", "lpUrlComponents"]),
        # 
        'WinHttpCreateUrl': SimTypeFunction([SimTypePointer(SimStruct({"dwStructSize": SimTypeInt(signed=False, label="UInt32"), "lpszScheme": SimTypePointer(SimTypeChar(label="Char"), offset=0), "dwSchemeLength": SimTypeInt(signed=False, label="UInt32"), "nScheme": SimTypeInt(signed=False, label="WINHTTP_INTERNET_SCHEME"), "lpszHostName": SimTypePointer(SimTypeChar(label="Char"), offset=0), "dwHostNameLength": SimTypeInt(signed=False, label="UInt32"), "nPort": SimTypeShort(signed=False, label="UInt16"), "lpszUserName": SimTypePointer(SimTypeChar(label="Char"), offset=0), "dwUserNameLength": SimTypeInt(signed=False, label="UInt32"), "lpszPassword": SimTypePointer(SimTypeChar(label="Char"), offset=0), "dwPasswordLength": SimTypeInt(signed=False, label="UInt32"), "lpszUrlPath": SimTypePointer(SimTypeChar(label="Char"), offset=0), "dwUrlPathLength": SimTypeInt(signed=False, label="UInt32"), "lpszExtraInfo": SimTypePointer(SimTypeChar(label="Char"), offset=0), "dwExtraInfoLength": SimTypeInt(signed=False, label="UInt32")}, name="URL_COMPONENTS", pack=False, align=None), offset=0), SimTypeInt(signed=False, label="WIN_HTTP_CREATE_URL_FLAGS"), SimTypePointer(SimTypeChar(label="Char"), label="LPArray", offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["lpUrlComponents", "dwFlags", "pwszUrl", "pdwUrlLength"]),
        # 
        'WinHttpCheckPlatform': SimTypeFunction([], SimTypeInt(signed=True, label="Int32")),
        # 
        'WinHttpGetDefaultProxyConfiguration': SimTypeFunction([SimTypePointer(SimStruct({"dwAccessType": SimTypeInt(signed=False, label="WINHTTP_ACCESS_TYPE"), "lpszProxy": SimTypePointer(SimTypeChar(label="Char"), offset=0), "lpszProxyBypass": SimTypePointer(SimTypeChar(label="Char"), offset=0)}, name="WINHTTP_PROXY_INFO", pack=False, align=None), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["pProxyInfo"]),
        # 
        'WinHttpSetDefaultProxyConfiguration': SimTypeFunction([SimTypePointer(SimStruct({"dwAccessType": SimTypeInt(signed=False, label="WINHTTP_ACCESS_TYPE"), "lpszProxy": SimTypePointer(SimTypeChar(label="Char"), offset=0), "lpszProxyBypass": SimTypePointer(SimTypeChar(label="Char"), offset=0)}, name="WINHTTP_PROXY_INFO", pack=False, align=None), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["pProxyInfo"]),
        # 
        'WinHttpOpen': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypeInt(signed=False, label="WINHTTP_ACCESS_TYPE"), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypeInt(signed=False, label="UInt32")], SimTypePointer(SimTypeBottom(label="Void"), offset=0), arg_names=["pszAgentW", "dwAccessType", "pszProxyW", "pszProxyBypassW", "dwFlags"]),
        # 
        'WinHttpCloseHandle': SimTypeFunction([SimTypePointer(SimTypeBottom(label="Void"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["hInternet"]),
        # 
        'WinHttpConnect': SimTypeFunction([SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypeInt(signed=False, label="INTERNET_PORT"), SimTypeInt(signed=False, label="UInt32")], SimTypePointer(SimTypeBottom(label="Void"), offset=0), arg_names=["hSession", "pswzServerName", "nServerPort", "dwReserved"]),
        # 
        'WinHttpReadData': SimTypeFunction([SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["hRequest", "lpBuffer", "dwNumberOfBytesToRead", "lpdwNumberOfBytesRead"]),
        # 
        'WinHttpWriteData': SimTypeFunction([SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["hRequest", "lpBuffer", "dwNumberOfBytesToWrite", "lpdwNumberOfBytesWritten"]),
        # 
        'WinHttpQueryDataAvailable': SimTypeFunction([SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["hRequest", "lpdwNumberOfBytesAvailable"]),
        # 
        'WinHttpQueryOption': SimTypeFunction([SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["hInternet", "dwOption", "lpBuffer", "lpdwBufferLength"]),
        # 
        'WinHttpSetOption': SimTypeFunction([SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeBottom(label="Void"), label="LPArray", offset=0), SimTypeInt(signed=False, label="UInt32")], SimTypeInt(signed=True, label="Int32"), arg_names=["hInternet", "dwOption", "lpBuffer", "dwBufferLength"]),
        # 
        'WinHttpSetTimeouts': SimTypeFunction([SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypeInt(signed=True, label="Int32"), SimTypeInt(signed=True, label="Int32"), SimTypeInt(signed=True, label="Int32"), SimTypeInt(signed=True, label="Int32")], SimTypeInt(signed=True, label="Int32"), arg_names=["hInternet", "nResolveTimeout", "nConnectTimeout", "nSendTimeout", "nReceiveTimeout"]),
        # 
        'WinHttpOpenRequest': SimTypeFunction([SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypePointer(SimTypeChar(label="Char"), offset=0), offset=0), SimTypeInt(signed=False, label="WINHTTP_OPEN_REQUEST_FLAGS")], SimTypePointer(SimTypeBottom(label="Void"), offset=0), arg_names=["hConnect", "pwszVerb", "pwszObjectName", "pwszVersion", "pwszReferrer", "ppwszAcceptTypes", "dwFlags"]),
        # 
        'WinHttpAddRequestHeaders': SimTypeFunction([SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypePointer(SimTypeChar(label="Char"), label="LPArray", offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32")], SimTypeInt(signed=True, label="Int32"), arg_names=["hRequest", "lpszHeaders", "dwHeadersLength", "dwModifiers"]),
        # 
        'WinHttpAddRequestHeadersEx': SimTypeFunction([SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypeLongLong(signed=False, label="UInt64"), SimTypeLongLong(signed=False, label="UInt64"), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimStruct({"Anonymous1": SimUnion({"pwszName": SimTypePointer(SimTypeChar(label="Char"), offset=0), "pszName": SimTypePointer(SimTypeChar(label="Byte"), offset=0)}, name="<anon>", label="None"), "Anonymous2": SimUnion({"pwszValue": SimTypePointer(SimTypeChar(label="Char"), offset=0), "pszValue": SimTypePointer(SimTypeChar(label="Byte"), offset=0)}, name="<anon>", label="None")}, name="WINHTTP_EXTENDED_HEADER", pack=False, align=None), label="LPArray", offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["hRequest", "dwModifiers", "ullFlags", "ullExtra", "cHeaders", "pHeaders"]),
        # 
        'WinHttpSendRequest': SimTypeFunction([SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypePointer(SimTypeChar(label="Char"), label="LPArray", offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeInt(signed=False, label="UInt"), label="UIntPtr", offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["hRequest", "lpszHeaders", "dwHeadersLength", "lpOptional", "dwOptionalLength", "dwTotalLength", "dwContext"]),
        # 
        'WinHttpSetCredentials': SimTypeFunction([SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeBottom(label="Void"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["hRequest", "AuthTargets", "AuthScheme", "pwszUserName", "pwszPassword", "pAuthParams"]),
        # 
        'WinHttpQueryAuthSchemes': SimTypeFunction([SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["hRequest", "lpdwSupportedSchemes", "lpdwFirstScheme", "pdwAuthTarget"]),
        # 
        'WinHttpReceiveResponse': SimTypeFunction([SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypePointer(SimTypeBottom(label="Void"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["hRequest", "lpReserved"]),
        # 
        'WinHttpQueryHeaders': SimTypeFunction([SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["hRequest", "dwInfoLevel", "pwszName", "lpBuffer", "lpdwBufferLength", "lpdwIndex"]),
        # 
        'WinHttpDetectAutoProxyConfigUrl': SimTypeFunction([SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypePointer(SimTypeChar(label="Char"), offset=0), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["dwAutoDetectFlags", "ppwstrAutoConfigUrl"]),
        # 
        'WinHttpGetProxyForUrl': SimTypeFunction([SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimStruct({"dwFlags": SimTypeInt(signed=False, label="UInt32"), "dwAutoDetectFlags": SimTypeInt(signed=False, label="UInt32"), "lpszAutoConfigUrl": SimTypePointer(SimTypeChar(label="Char"), offset=0), "lpvReserved": SimTypePointer(SimTypeBottom(label="Void"), offset=0), "dwReserved": SimTypeInt(signed=False, label="UInt32"), "fAutoLogonIfChallenged": SimTypeInt(signed=True, label="Int32")}, name="WINHTTP_AUTOPROXY_OPTIONS", pack=False, align=None), offset=0), SimTypePointer(SimStruct({"dwAccessType": SimTypeInt(signed=False, label="WINHTTP_ACCESS_TYPE"), "lpszProxy": SimTypePointer(SimTypeChar(label="Char"), offset=0), "lpszProxyBypass": SimTypePointer(SimTypeChar(label="Char"), offset=0)}, name="WINHTTP_PROXY_INFO", pack=False, align=None), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["hSession", "lpcwszUrl", "pAutoProxyOptions", "pProxyInfo"]),
        # 
        'WinHttpCreateProxyResolver': SimTypeFunction([SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypePointer(SimTypePointer(SimTypeBottom(label="Void"), offset=0), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["hSession", "phResolver"]),
        # 
        'WinHttpGetProxyForUrlEx': SimTypeFunction([SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimStruct({"dwFlags": SimTypeInt(signed=False, label="UInt32"), "dwAutoDetectFlags": SimTypeInt(signed=False, label="UInt32"), "lpszAutoConfigUrl": SimTypePointer(SimTypeChar(label="Char"), offset=0), "lpvReserved": SimTypePointer(SimTypeBottom(label="Void"), offset=0), "dwReserved": SimTypeInt(signed=False, label="UInt32"), "fAutoLogonIfChallenged": SimTypeInt(signed=True, label="Int32")}, name="WINHTTP_AUTOPROXY_OPTIONS", pack=False, align=None), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt"), label="UIntPtr", offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["hResolver", "pcwszUrl", "pAutoProxyOptions", "pContext"]),
        # 
        'WinHttpGetProxyForUrlEx2': SimTypeFunction([SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimStruct({"dwFlags": SimTypeInt(signed=False, label="UInt32"), "dwAutoDetectFlags": SimTypeInt(signed=False, label="UInt32"), "lpszAutoConfigUrl": SimTypePointer(SimTypeChar(label="Char"), offset=0), "lpvReserved": SimTypePointer(SimTypeBottom(label="Void"), offset=0), "dwReserved": SimTypeInt(signed=False, label="UInt32"), "fAutoLogonIfChallenged": SimTypeInt(signed=True, label="Int32")}, name="WINHTTP_AUTOPROXY_OPTIONS", pack=False, align=None), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeChar(label="Byte"), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt"), label="UIntPtr", offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["hResolver", "pcwszUrl", "pAutoProxyOptions", "cbInterfaceSelectionContext", "pInterfaceSelectionContext", "pContext"]),
        # 
        'WinHttpGetProxyResult': SimTypeFunction([SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypePointer(SimStruct({"cEntries": SimTypeInt(signed=False, label="UInt32"), "pEntries": SimTypePointer(SimStruct({"fProxy": SimTypeInt(signed=True, label="Int32"), "fBypass": SimTypeInt(signed=True, label="Int32"), "ProxyScheme": SimTypeInt(signed=False, label="WINHTTP_INTERNET_SCHEME"), "pwszProxy": SimTypePointer(SimTypeChar(label="Char"), offset=0), "ProxyPort": SimTypeShort(signed=False, label="UInt16")}, name="WINHTTP_PROXY_RESULT_ENTRY", pack=False, align=None), offset=0)}, name="WINHTTP_PROXY_RESULT", pack=False, align=None), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["hResolver", "pProxyResult"]),
        # 
        'WinHttpGetProxyResultEx': SimTypeFunction([SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypePointer(SimStruct({"cEntries": SimTypeInt(signed=False, label="UInt32"), "pEntries": SimTypePointer(SimStruct({"fProxy": SimTypeInt(signed=True, label="Int32"), "fBypass": SimTypeInt(signed=True, label="Int32"), "ProxyScheme": SimTypeInt(signed=False, label="WINHTTP_INTERNET_SCHEME"), "pwszProxy": SimTypePointer(SimTypeChar(label="Char"), offset=0), "ProxyPort": SimTypeShort(signed=False, label="UInt16")}, name="WINHTTP_PROXY_RESULT_ENTRY", pack=False, align=None), offset=0), "hProxyDetectionHandle": SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), "dwProxyInterfaceAffinity": SimTypeInt(signed=False, label="UInt32")}, name="WINHTTP_PROXY_RESULT_EX", pack=False, align=None), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["hResolver", "pProxyResultEx"]),
        # 
        'WinHttpFreeProxyResult': SimTypeFunction([SimTypePointer(SimStruct({"cEntries": SimTypeInt(signed=False, label="UInt32"), "pEntries": SimTypePointer(SimStruct({"fProxy": SimTypeInt(signed=True, label="Int32"), "fBypass": SimTypeInt(signed=True, label="Int32"), "ProxyScheme": SimTypeInt(signed=False, label="WINHTTP_INTERNET_SCHEME"), "pwszProxy": SimTypePointer(SimTypeChar(label="Char"), offset=0), "ProxyPort": SimTypeShort(signed=False, label="UInt16")}, name="WINHTTP_PROXY_RESULT_ENTRY", pack=False, align=None), offset=0)}, name="WINHTTP_PROXY_RESULT", pack=False, align=None), offset=0)], SimTypeBottom(label="Void"), arg_names=["pProxyResult"]),
        # 
        'WinHttpFreeProxyResultEx': SimTypeFunction([SimTypePointer(SimStruct({"cEntries": SimTypeInt(signed=False, label="UInt32"), "pEntries": SimTypePointer(SimStruct({"fProxy": SimTypeInt(signed=True, label="Int32"), "fBypass": SimTypeInt(signed=True, label="Int32"), "ProxyScheme": SimTypeInt(signed=False, label="WINHTTP_INTERNET_SCHEME"), "pwszProxy": SimTypePointer(SimTypeChar(label="Char"), offset=0), "ProxyPort": SimTypeShort(signed=False, label="UInt16")}, name="WINHTTP_PROXY_RESULT_ENTRY", pack=False, align=None), offset=0), "hProxyDetectionHandle": SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), "dwProxyInterfaceAffinity": SimTypeInt(signed=False, label="UInt32")}, name="WINHTTP_PROXY_RESULT_EX", pack=False, align=None), offset=0)], SimTypeBottom(label="Void"), arg_names=["pProxyResultEx"]),
        # 
        'WinHttpResetAutoProxy': SimTypeFunction([SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypeInt(signed=False, label="UInt32")], SimTypeInt(signed=False, label="UInt32"), arg_names=["hSession", "dwFlags"]),
        # 
        'WinHttpGetIEProxyConfigForCurrentUser': SimTypeFunction([SimTypePointer(SimStruct({"fAutoDetect": SimTypeInt(signed=True, label="Int32"), "lpszAutoConfigUrl": SimTypePointer(SimTypeChar(label="Char"), offset=0), "lpszProxy": SimTypePointer(SimTypeChar(label="Char"), offset=0), "lpszProxyBypass": SimTypePointer(SimTypeChar(label="Char"), offset=0)}, name="WINHTTP_CURRENT_USER_IE_PROXY_CONFIG", pack=False, align=None), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["pProxyConfig"]),
        # 
        'WinHttpWriteProxySettings': SimTypeFunction([SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypeInt(signed=True, label="Int32"), SimTypePointer(SimStruct({"dwStructSize": SimTypeInt(signed=False, label="UInt32"), "dwFlags": SimTypeInt(signed=False, label="UInt32"), "dwCurrentSettingsVersion": SimTypeInt(signed=False, label="UInt32"), "pwszConnectionName": SimTypePointer(SimTypeChar(label="Char"), offset=0), "pwszProxy": SimTypePointer(SimTypeChar(label="Char"), offset=0), "pwszProxyBypass": SimTypePointer(SimTypeChar(label="Char"), offset=0), "pwszAutoconfigUrl": SimTypePointer(SimTypeChar(label="Char"), offset=0), "pwszAutoconfigSecondaryUrl": SimTypePointer(SimTypeChar(label="Char"), offset=0), "dwAutoDiscoveryFlags": SimTypeInt(signed=False, label="UInt32"), "pwszLastKnownGoodAutoConfigUrl": SimTypePointer(SimTypeChar(label="Char"), offset=0), "dwAutoconfigReloadDelayMins": SimTypeInt(signed=False, label="UInt32"), "ftLastKnownDetectTime": SimStruct({"dwLowDateTime": SimTypeInt(signed=False, label="UInt32"), "dwHighDateTime": SimTypeInt(signed=False, label="UInt32")}, name="FILETIME", pack=False, align=None), "dwDetectedInterfaceIpCount": SimTypeInt(signed=False, label="UInt32"), "pdwDetectedInterfaceIp": SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0), "cNetworkKeys": SimTypeInt(signed=False, label="UInt32"), "pNetworkKeys": SimTypePointer(SimStruct({"pbBuffer": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 128)}, name="_WinHttpProxyNetworkKey", pack=False, align=None), offset=0)}, name="WINHTTP_PROXY_SETTINGS", pack=False, align=None), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["hSession", "fForceUpdate", "pWinHttpProxySettings"]),
        # 
        'WinHttpReadProxySettings': SimTypeFunction([SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypeInt(signed=True, label="Int32"), SimTypeInt(signed=True, label="Int32"), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0), SimTypePointer(SimTypeInt(signed=True, label="Int32"), offset=0), SimTypePointer(SimStruct({"dwStructSize": SimTypeInt(signed=False, label="UInt32"), "dwFlags": SimTypeInt(signed=False, label="UInt32"), "dwCurrentSettingsVersion": SimTypeInt(signed=False, label="UInt32"), "pwszConnectionName": SimTypePointer(SimTypeChar(label="Char"), offset=0), "pwszProxy": SimTypePointer(SimTypeChar(label="Char"), offset=0), "pwszProxyBypass": SimTypePointer(SimTypeChar(label="Char"), offset=0), "pwszAutoconfigUrl": SimTypePointer(SimTypeChar(label="Char"), offset=0), "pwszAutoconfigSecondaryUrl": SimTypePointer(SimTypeChar(label="Char"), offset=0), "dwAutoDiscoveryFlags": SimTypeInt(signed=False, label="UInt32"), "pwszLastKnownGoodAutoConfigUrl": SimTypePointer(SimTypeChar(label="Char"), offset=0), "dwAutoconfigReloadDelayMins": SimTypeInt(signed=False, label="UInt32"), "ftLastKnownDetectTime": SimStruct({"dwLowDateTime": SimTypeInt(signed=False, label="UInt32"), "dwHighDateTime": SimTypeInt(signed=False, label="UInt32")}, name="FILETIME", pack=False, align=None), "dwDetectedInterfaceIpCount": SimTypeInt(signed=False, label="UInt32"), "pdwDetectedInterfaceIp": SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0), "cNetworkKeys": SimTypeInt(signed=False, label="UInt32"), "pNetworkKeys": SimTypePointer(SimStruct({"pbBuffer": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 128)}, name="_WinHttpProxyNetworkKey", pack=False, align=None), offset=0)}, name="WINHTTP_PROXY_SETTINGS", pack=False, align=None), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["hSession", "pcwszConnectionName", "fFallBackToDefaultSettings", "fSetAutoDiscoverForDefaultSettings", "pdwSettingsVersion", "pfDefaultSettingsAreReturned", "pWinHttpProxySettings"]),
        # 
        'WinHttpFreeProxySettings': SimTypeFunction([SimTypePointer(SimStruct({"dwStructSize": SimTypeInt(signed=False, label="UInt32"), "dwFlags": SimTypeInt(signed=False, label="UInt32"), "dwCurrentSettingsVersion": SimTypeInt(signed=False, label="UInt32"), "pwszConnectionName": SimTypePointer(SimTypeChar(label="Char"), offset=0), "pwszProxy": SimTypePointer(SimTypeChar(label="Char"), offset=0), "pwszProxyBypass": SimTypePointer(SimTypeChar(label="Char"), offset=0), "pwszAutoconfigUrl": SimTypePointer(SimTypeChar(label="Char"), offset=0), "pwszAutoconfigSecondaryUrl": SimTypePointer(SimTypeChar(label="Char"), offset=0), "dwAutoDiscoveryFlags": SimTypeInt(signed=False, label="UInt32"), "pwszLastKnownGoodAutoConfigUrl": SimTypePointer(SimTypeChar(label="Char"), offset=0), "dwAutoconfigReloadDelayMins": SimTypeInt(signed=False, label="UInt32"), "ftLastKnownDetectTime": SimStruct({"dwLowDateTime": SimTypeInt(signed=False, label="UInt32"), "dwHighDateTime": SimTypeInt(signed=False, label="UInt32")}, name="FILETIME", pack=False, align=None), "dwDetectedInterfaceIpCount": SimTypeInt(signed=False, label="UInt32"), "pdwDetectedInterfaceIp": SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0), "cNetworkKeys": SimTypeInt(signed=False, label="UInt32"), "pNetworkKeys": SimTypePointer(SimStruct({"pbBuffer": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 128)}, name="_WinHttpProxyNetworkKey", pack=False, align=None), offset=0)}, name="WINHTTP_PROXY_SETTINGS", pack=False, align=None), offset=0)], SimTypeBottom(label="Void"), arg_names=["pWinHttpProxySettings"]),
        # 
        'WinHttpGetProxySettingsVersion': SimTypeFunction([SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["hSession", "pdwProxySettingsVersion"]),
        # 
        'WinHttpSetProxySettingsPerUser': SimTypeFunction([SimTypeInt(signed=True, label="Int32")], SimTypeInt(signed=False, label="UInt32"), arg_names=["fProxySettingsPerUser"]),
        # 
        'WinHttpWebSocketCompleteUpgrade': SimTypeFunction([SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt"), label="UIntPtr", offset=0)], SimTypePointer(SimTypeBottom(label="Void"), offset=0), arg_names=["hRequest", "pContext"]),
        # 
        'WinHttpWebSocketSend': SimTypeFunction([SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypeInt(signed=False, label="WINHTTP_WEB_SOCKET_BUFFER_TYPE"), SimTypePointer(SimTypeBottom(label="Void"), label="LPArray", offset=0), SimTypeInt(signed=False, label="UInt32")], SimTypeInt(signed=False, label="UInt32"), arg_names=["hWebSocket", "eBufferType", "pvBuffer", "dwBufferLength"]),
        # 
        'WinHttpWebSocketReceive': SimTypeFunction([SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0), SimTypePointer(SimTypeInt(signed=False, label="WINHTTP_WEB_SOCKET_BUFFER_TYPE"), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["hWebSocket", "pvBuffer", "dwBufferLength", "pdwBytesRead", "peBufferType"]),
        # 
        'WinHttpWebSocketShutdown': SimTypeFunction([SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypeShort(signed=False, label="UInt16"), SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypeInt(signed=False, label="UInt32")], SimTypeInt(signed=False, label="UInt32"), arg_names=["hWebSocket", "usStatus", "pvReason", "dwReasonLength"]),
        # 
        'WinHttpWebSocketClose': SimTypeFunction([SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypeShort(signed=False, label="UInt16"), SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypeInt(signed=False, label="UInt32")], SimTypeInt(signed=False, label="UInt32"), arg_names=["hWebSocket", "usStatus", "pvReason", "dwReasonLength"]),
        # 
        'WinHttpWebSocketQueryCloseStatus': SimTypeFunction([SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypePointer(SimTypeShort(signed=False, label="UInt16"), offset=0), SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["hWebSocket", "pusStatus", "pvReason", "dwReasonLength", "pdwReasonLengthConsumed"]),
    }

lib.set_prototypes(prototypes)



from .base import BaseStructuredCodeGenerator, InstructionMapping, InstructionMappingElement, PositionMappingElement,\
                  PositionMapping
from .c import CStructuredCodeGenerator
from .dwarf_import import ImportSourceCode
from .dummy import DummyStructuredCodeGenerator



###
### This file was automatically generated
###

from archinfo.arch import register_arch, Endness, Register

from .common import ArchPcode


class ArchPcode_MIPS_LE_64_64_32R6addr(ArchPcode):
    name = 'MIPS:LE:64:64-32R6addr'
    pcode_arch = 'MIPS:LE:64:64-32R6addr'
    description = 'MIPS64 Release-6 with 32-bit addresses, little endian, with microMIPS'
    bits = 32
    ip_offset = 0x100
    sp_offset = 0xe8
    bp_offset = sp_offset
    instruction_endness = Endness.LE
    register_list = [
        Register('zero', 8, 0x0),
        Register('zero_lo', 4, 0x0),
        Register('zero_hi', 4, 0x4),
        Register('at', 8, 0x8),
        Register('at_lo', 4, 0x8),
        Register('at_hi', 4, 0xc),
        Register('v0', 8, 0x10),
        Register('v0_lo', 4, 0x10),
        Register('v0_hi', 4, 0x14),
        Register('v1', 8, 0x18),
        Register('v1_lo', 4, 0x18),
        Register('v1_hi', 4, 0x1c),
        Register('a0', 8, 0x20),
        Register('a0_lo', 4, 0x20),
        Register('a0_hi', 4, 0x24),
        Register('a1', 8, 0x28),
        Register('a1_lo', 4, 0x28),
        Register('a1_hi', 4, 0x2c),
        Register('a2', 8, 0x30),
        Register('a2_lo', 4, 0x30),
        Register('a2_hi', 4, 0x34),
        Register('a3', 8, 0x38),
        Register('a3_lo', 4, 0x38),
        Register('a3_hi', 4, 0x3c),
        Register('t0', 8, 0x40),
        Register('t0_lo', 4, 0x40),
        Register('t0_hi', 4, 0x44),
        Register('t1', 8, 0x48),
        Register('t1_lo', 4, 0x48),
        Register('t1_hi', 4, 0x4c),
        Register('t2', 8, 0x50),
        Register('t2_lo', 4, 0x50),
        Register('t2_hi', 4, 0x54),
        Register('t3', 8, 0x58),
        Register('t3_lo', 4, 0x58),
        Register('t3_hi', 4, 0x5c),
        Register('t4', 8, 0x60),
        Register('t4_lo', 4, 0x60),
        Register('t4_hi', 4, 0x64),
        Register('t5', 8, 0x68),
        Register('t5_lo', 4, 0x68),
        Register('t5_hi', 4, 0x6c),
        Register('t6', 8, 0x70),
        Register('t6_lo', 4, 0x70),
        Register('t6_hi', 4, 0x74),
        Register('t7', 8, 0x78),
        Register('t7_lo', 4, 0x78),
        Register('t7_hi', 4, 0x7c),
        Register('s0', 8, 0x80),
        Register('s0_lo', 4, 0x80),
        Register('s0_hi', 4, 0x84),
        Register('s1', 8, 0x88),
        Register('s1_lo', 4, 0x88),
        Register('s1_hi', 4, 0x8c),
        Register('s2', 8, 0x90),
        Register('s2_lo', 4, 0x90),
        Register('s2_hi', 4, 0x94),
        Register('s3', 8, 0x98),
        Register('s3_lo', 4, 0x98),
        Register('s3_hi', 4, 0x9c),
        Register('s4', 8, 0xa0),
        Register('s4_lo', 4, 0xa0),
        Register('s4_hi', 4, 0xa4),
        Register('s5', 8, 0xa8),
        Register('s5_lo', 4, 0xa8),
        Register('s5_hi', 4, 0xac),
        Register('s6', 8, 0xb0),
        Register('s6_lo', 4, 0xb0),
        Register('s6_hi', 4, 0xb4),
        Register('s7', 8, 0xb8),
        Register('s7_lo', 4, 0xb8),
        Register('s7_hi', 4, 0xbc),
        Register('t8', 8, 0xc0),
        Register('t8_lo', 4, 0xc0),
        Register('t8_hi', 4, 0xc4),
        Register('t9', 8, 0xc8),
        Register('t9_lo', 4, 0xc8),
        Register('t9_hi', 4, 0xcc),
        Register('k0', 8, 0xd0),
        Register('k0_lo', 4, 0xd0),
        Register('k0_hi', 4, 0xd4),
        Register('k1', 8, 0xd8),
        Register('k1_lo', 4, 0xd8),
        Register('k1_hi', 4, 0xdc),
        Register('gp', 8, 0xe0),
        Register('gp_lo', 4, 0xe0),
        Register('gp_hi', 4, 0xe4),
        Register('sp', 8, 0xe8),
        Register('sp_lo', 4, 0xe8),
        Register('sp_hi', 4, 0xec),
        Register('s8', 8, 0xf0),
        Register('s8_lo', 4, 0xf0),
        Register('s8_hi', 4, 0xf4),
        Register('ra', 8, 0xf8),
        Register('ra_lo', 4, 0xf8),
        Register('ra_hi', 4, 0xfc),
        Register('pc', 8, 0x100, alias_names=('ip',)),
        Register('pc_lo', 4, 0x100),
        Register('pc_hi', 4, 0x104),
        Register('f0', 8, 0x1000),
        Register('f1', 8, 0x1008),
        Register('f2', 8, 0x1010),
        Register('f3', 8, 0x1018),
        Register('f4', 8, 0x1020),
        Register('f5', 8, 0x1028),
        Register('f6', 8, 0x1030),
        Register('f7', 8, 0x1038),
        Register('f8', 8, 0x1040),
        Register('f9', 8, 0x1048),
        Register('f10', 8, 0x1050),
        Register('f11', 8, 0x1058),
        Register('f12', 8, 0x1060),
        Register('f13', 8, 0x1068),
        Register('f14', 8, 0x1070),
        Register('f15', 8, 0x1078),
        Register('f16', 8, 0x1080),
        Register('f17', 8, 0x1088),
        Register('f18', 8, 0x1090),
        Register('f19', 8, 0x1098),
        Register('f20', 8, 0x10a0),
        Register('f21', 8, 0x10a8),
        Register('f22', 8, 0x10b0),
        Register('f23', 8, 0x10b8),
        Register('f24', 8, 0x10c0),
        Register('f25', 8, 0x10c8),
        Register('f26', 8, 0x10d0),
        Register('f27', 8, 0x10d8),
        Register('f28', 8, 0x10e0),
        Register('f29', 8, 0x10e8),
        Register('f30', 8, 0x10f0),
        Register('f31', 8, 0x10f8),
        Register('fir', 4, 0x1200),
        Register('fccr', 4, 0x1204),
        Register('fexr', 4, 0x1208),
        Register('fenr', 4, 0x120c),
        Register('fcsr', 4, 0x1210),
        Register('index', 8, 0x2000),
        Register('random', 8, 0x2008),
        Register('entrylo0', 8, 0x2010),
        Register('entrylo1', 8, 0x2018),
        Register('context', 8, 0x2020),
        Register('pagemask', 8, 0x2028),
        Register('wired', 8, 0x2030),
        Register('hwrena', 8, 0x2038),
        Register('badvaddr', 8, 0x2040),
        Register('count', 8, 0x2048),
        Register('entryhi', 8, 0x2050),
        Register('compare', 8, 0x2058),
        Register('status', 8, 0x2060),
        Register('cause', 8, 0x2068),
        Register('epc', 8, 0x2070),
        Register('prid', 8, 0x2078),
        Register('config', 8, 0x2080),
        Register('lladdr', 8, 0x2088),
        Register('watchlo', 8, 0x2090),
        Register('watchhi', 8, 0x2098),
        Register('xcontext', 8, 0x20a0),
        Register('cop0_reg21', 8, 0x20a8),
        Register('cop0_reg22', 8, 0x20b0),
        Register('debug', 8, 0x20b8),
        Register('depc', 8, 0x20c0),
        Register('perfcnt', 8, 0x20c8),
        Register('errctl', 8, 0x20d0),
        Register('cacheerr', 8, 0x20d8),
        Register('taglo', 8, 0x20e0),
        Register('taghi', 8, 0x20e8),
        Register('errorepc', 8, 0x20f0),
        Register('desave', 8, 0x20f8),
        Register('mvpcontrol', 8, 0x2100),
        Register('vpecontrol', 8, 0x2108),
        Register('tcstatus', 8, 0x2110),
        Register('cop0_reg3.1', 8, 0x2118),
        Register('contextconfig', 8, 0x2120),
        Register('pagegrain', 8, 0x2128),
        Register('srsconf0', 8, 0x2130),
        Register('cop0_reg7.1', 8, 0x2138),
        Register('cop0_reg8.1', 8, 0x2140),
        Register('cop0_reg9.1', 8, 0x2148),
        Register('cop0_reg10.1', 8, 0x2150),
        Register('cop0_reg11.1', 8, 0x2158),
        Register('intctl', 8, 0x2160),
        Register('cop0_reg13.1', 8, 0x2168),
        Register('cop0_reg14.1', 8, 0x2170),
        Register('ebase', 8, 0x2178),
        Register('config1', 8, 0x2180),
        Register('cop0_reg17.1', 8, 0x2188),
        Register('watchlo.1', 8, 0x2190),
        Register('watchhi.1', 8, 0x2198),
        Register('cop0_reg20.1', 8, 0x21a0),
        Register('cop0_reg21.1', 8, 0x21a8),
        Register('cop0_reg22.1', 8, 0x21b0),
        Register('tracecontrol', 8, 0x21b8),
        Register('cop0_reg24.1', 8, 0x21c0),
        Register('perfcnt.1', 8, 0x21c8),
        Register('cop0_reg26.1', 8, 0x21d0),
        Register('cacheerr.1', 8, 0x21d8),
        Register('datalo.1', 8, 0x21e0),
        Register('datahi.1', 8, 0x21e8),
        Register('cop0_reg30.1', 8, 0x21f0),
        Register('cop0_reg31.1', 8, 0x21f8),
        Register('mvpconf0', 8, 0x2200),
        Register('vpeconf0', 8, 0x2208),
        Register('tcbind', 8, 0x2210),
        Register('cop0_reg3.2', 8, 0x2218),
        Register('cop0_reg4.2', 8, 0x2220),
        Register('cop0_reg5.2', 8, 0x2228),
        Register('srsconf1', 8, 0x2230),
        Register('cop0_reg7.2', 8, 0x2238),
        Register('cop0_reg8.2', 8, 0x2240),
        Register('cop0_reg9.2', 8, 0x2248),
        Register('cop0_reg10.2', 8, 0x2250),
        Register('cop0_reg11.2', 8, 0x2258),
        Register('srsctl', 8, 0x2260),
        Register('cop0_reg13.2', 8, 0x2268),
        Register('cop0_reg14.2', 8, 0x2270),
        Register('cop0_reg15.2', 8, 0x2278),
        Register('config2', 8, 0x2280),
        Register('cop0_reg17.2', 8, 0x2288),
        Register('watchlo.2', 8, 0x2290),
        Register('watchhi.2', 8, 0x2298),
        Register('cop0_reg20.2', 8, 0x22a0),
        Register('cop0_reg21.2', 8, 0x22a8),
        Register('cop0_reg22.2', 8, 0x22b0),
        Register('tracecontrol2', 8, 0x22b8),
        Register('cop0_reg24.2', 8, 0x22c0),
        Register('perfcnt.2', 8, 0x22c8),
        Register('cop0_reg26.2', 8, 0x22d0),
        Register('cacheerr.2', 8, 0x22d8),
        Register('taglo.2', 8, 0x22e0),
        Register('taghi.2', 8, 0x22e8),
        Register('cop0_reg30.2', 8, 0x22f0),
        Register('cop0_reg31.2', 8, 0x22f8),
        Register('mvpconf1', 8, 0x2300),
        Register('vpeconf1', 8, 0x2308),
        Register('tcrestart', 8, 0x2310),
        Register('cop0_reg3.3', 8, 0x2318),
        Register('cop0_reg4.3', 8, 0x2320),
        Register('cop0_reg5.3', 8, 0x2328),
        Register('srsconf2', 8, 0x2330),
        Register('cop0_reg7.3', 8, 0x2338),
        Register('cop0_reg8.3', 8, 0x2340),
        Register('cop0_reg9.3', 8, 0x2348),
        Register('cop0_reg10.3', 8, 0x2350),
        Register('cop0_reg11.3', 8, 0x2358),
        Register('srsmap', 8, 0x2360),
        Register('cop0_reg13.3', 8, 0x2368),
        Register('cop0_reg14.3', 8, 0x2370),
        Register('cop0_reg15.3', 8, 0x2378),
        Register('config3', 8, 0x2380),
        Register('cop0_reg17.3', 8, 0x2388),
        Register('watchlo.3', 8, 0x2390),
        Register('watchhi.3', 8, 0x2398),
        Register('cop0_reg20.3', 8, 0x23a0),
        Register('cop0_reg21.3', 8, 0x23a8),
        Register('cop0_reg22.3', 8, 0x23b0),
        Register('usertracedata', 8, 0x23b8),
        Register('cop0_reg24.3', 8, 0x23c0),
        Register('perfcnt.3', 8, 0x23c8),
        Register('cop0_reg26.3', 8, 0x23d0),
        Register('cacheerr.3', 8, 0x23d8),
        Register('datalo.3', 8, 0x23e0),
        Register('datahi.3', 8, 0x23e8),
        Register('cop0_reg30.3', 8, 0x23f0),
        Register('cop0_reg31.3', 8, 0x23f8),
        Register('cop0_reg0.4', 8, 0x2400),
        Register('yqmask', 8, 0x2408),
        Register('tchalt', 8, 0x2410),
        Register('cop0_reg3.4', 8, 0x2418),
        Register('cop0_reg4.4', 8, 0x2420),
        Register('cop0_reg5.4', 8, 0x2428),
        Register('srsconf3', 8, 0x2430),
        Register('cop0_reg7.4', 8, 0x2438),
        Register('cop0_reg8.4', 8, 0x2440),
        Register('cop0_reg9.4', 8, 0x2448),
        Register('cop0_reg10.4', 8, 0x2450),
        Register('cop0_reg11.4', 8, 0x2458),
        Register('cop0_reg12.4', 8, 0x2460),
        Register('cop0_reg13.4', 8, 0x2468),
        Register('cop0_reg14.4', 8, 0x2470),
        Register('cop0_reg15.4', 8, 0x2478),
        Register('cop0_reg16.4', 8, 0x2480),
        Register('cop0_reg17.4', 8, 0x2488),
        Register('watchlo.4', 8, 0x2490),
        Register('watchhi.4', 8, 0x2498),
        Register('cop0_reg20.4', 8, 0x24a0),
        Register('cop0_reg21.4', 8, 0x24a8),
        Register('cop0_reg22.4', 8, 0x24b0),
        Register('tracebpc', 8, 0x24b8),
        Register('cop0_reg24.4', 8, 0x24c0),
        Register('perfcnt.4', 8, 0x24c8),
        Register('cop0_reg26.4', 8, 0x24d0),
        Register('cacheerr.4', 8, 0x24d8),
        Register('taglo.4', 8, 0x24e0),
        Register('taghi.4', 8, 0x24e8),
        Register('cop0_reg30.4', 8, 0x24f0),
        Register('cop0_reg31.4', 8, 0x24f8),
        Register('cop0_reg0.5', 8, 0x2500),
        Register('vpeschedule', 8, 0x2508),
        Register('tccontext', 8, 0x2510),
        Register('cop0_reg3.5', 8, 0x2518),
        Register('cop0_reg4.5', 8, 0x2520),
        Register('cop0_reg5.5', 8, 0x2528),
        Register('srsconf4', 8, 0x2530),
        Register('cop0_reg7.5', 8, 0x2538),
        Register('cop0_reg8.5', 8, 0x2540),
        Register('cop0_reg9.5', 8, 0x2548),
        Register('cop0_reg10.5', 8, 0x2550),
        Register('cop0_reg11.5', 8, 0x2558),
        Register('cop0_reg12.5', 8, 0x2560),
        Register('cop0_reg13.5', 8, 0x2568),
        Register('cop0_reg14.5', 8, 0x2570),
        Register('cop0_reg15.5', 8, 0x2578),
        Register('cop0_reg16.5', 8, 0x2580),
        Register('cop0_reg17.5', 8, 0x2588),
        Register('watchlo.5', 8, 0x2590),
        Register('watchhi.5', 8, 0x2598),
        Register('cop0_reg20.5', 8, 0x25a0),
        Register('cop0_reg21.5', 8, 0x25a8),
        Register('cop0_reg22.5', 8, 0x25b0),
        Register('cop0_reg23.5', 8, 0x25b8),
        Register('cop0_reg24.5', 8, 0x25c0),
        Register('perfcnt.5', 8, 0x25c8),
        Register('cop0_reg26.5', 8, 0x25d0),
        Register('cacheerr.5', 8, 0x25d8),
        Register('datalo.5', 8, 0x25e0),
        Register('datahi.5', 8, 0x25e8),
        Register('cop0_reg30.5', 8, 0x25f0),
        Register('cop0_reg31.5', 8, 0x25f8),
        Register('cop0_reg0.6', 8, 0x2600),
        Register('vpeschefback', 8, 0x2608),
        Register('tcschedule', 8, 0x2610),
        Register('cop0_reg3.6', 8, 0x2618),
        Register('cop0_reg4.6', 8, 0x2620),
        Register('cop0_reg5.6', 8, 0x2628),
        Register('cop0_reg6.6', 8, 0x2630),
        Register('cop0_reg7.6', 8, 0x2638),
        Register('cop0_reg8.6', 8, 0x2640),
        Register('cop0_reg9.6', 8, 0x2648),
        Register('cop0_reg10.6', 8, 0x2650),
        Register('cop0_reg11.6', 8, 0x2658),
        Register('cop0_reg12.6', 8, 0x2660),
        Register('cop0_reg13.6', 8, 0x2668),
        Register('cop0_reg14.6', 8, 0x2670),
        Register('cop0_reg15.6', 8, 0x2678),
        Register('cop0_reg16.6', 8, 0x2680),
        Register('cop0_reg17.6', 8, 0x2688),
        Register('watchlo.6', 8, 0x2690),
        Register('watchhi.6', 8, 0x2698),
        Register('cop0_reg20.6', 8, 0x26a0),
        Register('cop0_reg21.6', 8, 0x26a8),
        Register('cop0_reg22.6', 8, 0x26b0),
        Register('cop0_reg23.6', 8, 0x26b8),
        Register('cop0_reg24.6', 8, 0x26c0),
        Register('perfcnt.6', 8, 0x26c8),
        Register('cop0_reg26.6', 8, 0x26d0),
        Register('cacheerr.6', 8, 0x26d8),
        Register('taglo.6', 8, 0x26e0),
        Register('taghi.6', 8, 0x26e8),
        Register('cop0_reg30.6', 8, 0x26f0),
        Register('cop0_reg31.6', 8, 0x26f8),
        Register('cop0_reg0.7', 8, 0x2700),
        Register('vpeopt', 8, 0x2708),
        Register('tcschefback', 8, 0x2710),
        Register('cop0_reg3.7', 8, 0x2718),
        Register('cop0_reg4.7', 8, 0x2720),
        Register('cop0_reg5.7', 8, 0x2728),
        Register('cop0_reg6.7', 8, 0x2730),
        Register('cop0_reg7.7', 8, 0x2738),
        Register('cop0_reg8.7', 8, 0x2740),
        Register('cop0_reg9.7', 8, 0x2748),
        Register('cop0_reg10.7', 8, 0x2750),
        Register('cop0_reg11.7', 8, 0x2758),
        Register('cop0_reg12.7', 8, 0x2760),
        Register('cop0_reg13.7', 8, 0x2768),
        Register('cop0_reg14.7', 8, 0x2770),
        Register('cop0_reg15.7', 8, 0x2778),
        Register('cop0_reg16.7', 8, 0x2780),
        Register('cop0_reg17.7', 8, 0x2788),
        Register('watchlo.7', 8, 0x2790),
        Register('watchhi.7', 8, 0x2798),
        Register('cop0_reg20.7', 8, 0x27a0),
        Register('cop0_reg21.7', 8, 0x27a8),
        Register('cop0_reg22.7', 8, 0x27b0),
        Register('cop0_reg23.7', 8, 0x27b8),
        Register('cop0_reg24.7', 8, 0x27c0),
        Register('perfcnt.7', 8, 0x27c8),
        Register('cop0_reg26.7', 8, 0x27d0),
        Register('cacheerr.7', 8, 0x27d8),
        Register('datalo.7', 8, 0x27e0),
        Register('datahi.7', 8, 0x27e8),
        Register('cop0_reg30.7', 8, 0x27f0),
        Register('cop0_reg31.7', 8, 0x27f8),
        Register('hi', 8, 0x3000),
        Register('lo', 8, 0x3008),
        Register('hi1', 8, 0x3010),
        Register('lo1', 8, 0x3018),
        Register('hi2', 8, 0x3020),
        Register('lo2', 8, 0x3028),
        Register('hi3', 8, 0x3030),
        Register('lo3', 8, 0x3038),
        Register('tsp', 8, 0x3040),
        Register('isamodeswitch', 1, 0x3f00),
        Register('contextreg', 4, 0x4000)
    ]

register_arch(['mips:le:64:64-32r6addr'], 32, Endness.LE, ArchPcode_MIPS_LE_64_64_32R6addr)



# pylint:disable=line-too-long
import logging

from ...sim_type import SimTypeFunction,     SimTypeShort, SimTypeInt, SimTypeLong, SimTypeLongLong, SimTypeDouble, SimTypeFloat,     SimTypePointer,     SimTypeChar,     SimStruct,     SimTypeFixedSizeArray,     SimTypeBottom,     SimUnion,     SimTypeBool
from ...calling_conventions import SimCCStdcall, SimCCMicrosoftAMD64
from .. import SIM_PROCEDURES as P
from . import SimLibrary


_l = logging.getLogger(name=__name__)


lib = SimLibrary()
lib.set_default_cc('X86', SimCCStdcall)
lib.set_default_cc('AMD64', SimCCMicrosoftAMD64)
lib.add('RtlEncodePointer', P['win32']['EncodePointer'])
lib.add('RtlDecodePointer', P['win32']['EncodePointer'])
lib.add('RtlAllocateHeap', P['win32']['HeapAlloc'])
lib.set_library_names("ntdll.dll")
prototypes = \
    {
        # 
        'RtlNtStatusToDosError': SimTypeFunction([SimTypeInt(signed=True, label="Int32")], SimTypeInt(signed=False, label="UInt32"), arg_names=["Status"]),
        # 
        'RtlIpv4AddressToStringA': SimTypeFunction([SimTypePointer(SimStruct({"S_un": SimUnion({"S_un_b": SimStruct({"s_b1": SimTypeChar(label="Byte"), "s_b2": SimTypeChar(label="Byte"), "s_b3": SimTypeChar(label="Byte"), "s_b4": SimTypeChar(label="Byte")}, name="_S_un_b_e__Struct", pack=False, align=None), "S_un_w": SimStruct({"s_w1": SimTypeShort(signed=False, label="UInt16"), "s_w2": SimTypeShort(signed=False, label="UInt16")}, name="_S_un_w_e__Struct", pack=False, align=None), "S_addr": SimTypeInt(signed=False, label="UInt32")}, name="<anon>", label="None")}, name="IN_ADDR", pack=False, align=None), offset=0), SimTypePointer(SimTypeChar(label="Byte"), label="LPArray", offset=0)], SimTypePointer(SimTypeChar(label="Byte"), offset=0), arg_names=["Addr", "S"]),
        # 
        'RtlIpv4AddressToStringExA': SimTypeFunction([SimTypePointer(SimStruct({"S_un": SimUnion({"S_un_b": SimStruct({"s_b1": SimTypeChar(label="Byte"), "s_b2": SimTypeChar(label="Byte"), "s_b3": SimTypeChar(label="Byte"), "s_b4": SimTypeChar(label="Byte")}, name="_S_un_b_e__Struct", pack=False, align=None), "S_un_w": SimStruct({"s_w1": SimTypeShort(signed=False, label="UInt16"), "s_w2": SimTypeShort(signed=False, label="UInt16")}, name="_S_un_w_e__Struct", pack=False, align=None), "S_addr": SimTypeInt(signed=False, label="UInt32")}, name="<anon>", label="None")}, name="IN_ADDR", pack=False, align=None), offset=0), SimTypeShort(signed=False, label="UInt16"), SimTypePointer(SimTypeChar(label="Byte"), label="LPArray", offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["Address", "Port", "AddressString", "AddressStringLength"]),
        # 
        'RtlIpv4AddressToStringW': SimTypeFunction([SimTypePointer(SimStruct({"S_un": SimUnion({"S_un_b": SimStruct({"s_b1": SimTypeChar(label="Byte"), "s_b2": SimTypeChar(label="Byte"), "s_b3": SimTypeChar(label="Byte"), "s_b4": SimTypeChar(label="Byte")}, name="_S_un_b_e__Struct", pack=False, align=None), "S_un_w": SimStruct({"s_w1": SimTypeShort(signed=False, label="UInt16"), "s_w2": SimTypeShort(signed=False, label="UInt16")}, name="_S_un_w_e__Struct", pack=False, align=None), "S_addr": SimTypeInt(signed=False, label="UInt32")}, name="<anon>", label="None")}, name="IN_ADDR", pack=False, align=None), offset=0), SimTypePointer(SimTypeChar(label="Char"), label="LPArray", offset=0)], SimTypePointer(SimTypeChar(label="Char"), offset=0), arg_names=["Addr", "S"]),
        # 
        'RtlIpv4AddressToStringExW': SimTypeFunction([SimTypePointer(SimStruct({"S_un": SimUnion({"S_un_b": SimStruct({"s_b1": SimTypeChar(label="Byte"), "s_b2": SimTypeChar(label="Byte"), "s_b3": SimTypeChar(label="Byte"), "s_b4": SimTypeChar(label="Byte")}, name="_S_un_b_e__Struct", pack=False, align=None), "S_un_w": SimStruct({"s_w1": SimTypeShort(signed=False, label="UInt16"), "s_w2": SimTypeShort(signed=False, label="UInt16")}, name="_S_un_w_e__Struct", pack=False, align=None), "S_addr": SimTypeInt(signed=False, label="UInt32")}, name="<anon>", label="None")}, name="IN_ADDR", pack=False, align=None), offset=0), SimTypeShort(signed=False, label="UInt16"), SimTypePointer(SimTypeChar(label="Char"), label="LPArray", offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["Address", "Port", "AddressString", "AddressStringLength"]),
        # 
        'RtlIpv4StringToAddressA': SimTypeFunction([SimTypePointer(SimTypeChar(label="Byte"), offset=0), SimTypeChar(label="Byte"), SimTypePointer(SimTypePointer(SimTypeChar(label="Byte"), offset=0), offset=0), SimTypePointer(SimStruct({"S_un": SimUnion({"S_un_b": SimStruct({"s_b1": SimTypeChar(label="Byte"), "s_b2": SimTypeChar(label="Byte"), "s_b3": SimTypeChar(label="Byte"), "s_b4": SimTypeChar(label="Byte")}, name="_S_un_b_e__Struct", pack=False, align=None), "S_un_w": SimStruct({"s_w1": SimTypeShort(signed=False, label="UInt16"), "s_w2": SimTypeShort(signed=False, label="UInt16")}, name="_S_un_w_e__Struct", pack=False, align=None), "S_addr": SimTypeInt(signed=False, label="UInt32")}, name="<anon>", label="None")}, name="IN_ADDR", pack=False, align=None), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["S", "Strict", "Terminator", "Addr"]),
        # 
        'RtlIpv4StringToAddressExA': SimTypeFunction([SimTypePointer(SimTypeChar(label="Byte"), offset=0), SimTypeChar(label="Byte"), SimTypePointer(SimStruct({"S_un": SimUnion({"S_un_b": SimStruct({"s_b1": SimTypeChar(label="Byte"), "s_b2": SimTypeChar(label="Byte"), "s_b3": SimTypeChar(label="Byte"), "s_b4": SimTypeChar(label="Byte")}, name="_S_un_b_e__Struct", pack=False, align=None), "S_un_w": SimStruct({"s_w1": SimTypeShort(signed=False, label="UInt16"), "s_w2": SimTypeShort(signed=False, label="UInt16")}, name="_S_un_w_e__Struct", pack=False, align=None), "S_addr": SimTypeInt(signed=False, label="UInt32")}, name="<anon>", label="None")}, name="IN_ADDR", pack=False, align=None), offset=0), SimTypePointer(SimTypeShort(signed=False, label="UInt16"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["AddressString", "Strict", "Address", "Port"]),
        # 
        'RtlIpv4StringToAddressW': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypeChar(label="Byte"), SimTypePointer(SimTypePointer(SimTypeChar(label="Char"), offset=0), offset=0), SimTypePointer(SimStruct({"S_un": SimUnion({"S_un_b": SimStruct({"s_b1": SimTypeChar(label="Byte"), "s_b2": SimTypeChar(label="Byte"), "s_b3": SimTypeChar(label="Byte"), "s_b4": SimTypeChar(label="Byte")}, name="_S_un_b_e__Struct", pack=False, align=None), "S_un_w": SimStruct({"s_w1": SimTypeShort(signed=False, label="UInt16"), "s_w2": SimTypeShort(signed=False, label="UInt16")}, name="_S_un_w_e__Struct", pack=False, align=None), "S_addr": SimTypeInt(signed=False, label="UInt32")}, name="<anon>", label="None")}, name="IN_ADDR", pack=False, align=None), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["S", "Strict", "Terminator", "Addr"]),
        # 
        'RtlIpv4StringToAddressExW': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypeChar(label="Byte"), SimTypePointer(SimStruct({"S_un": SimUnion({"S_un_b": SimStruct({"s_b1": SimTypeChar(label="Byte"), "s_b2": SimTypeChar(label="Byte"), "s_b3": SimTypeChar(label="Byte"), "s_b4": SimTypeChar(label="Byte")}, name="_S_un_b_e__Struct", pack=False, align=None), "S_un_w": SimStruct({"s_w1": SimTypeShort(signed=False, label="UInt16"), "s_w2": SimTypeShort(signed=False, label="UInt16")}, name="_S_un_w_e__Struct", pack=False, align=None), "S_addr": SimTypeInt(signed=False, label="UInt32")}, name="<anon>", label="None")}, name="IN_ADDR", pack=False, align=None), offset=0), SimTypePointer(SimTypeShort(signed=False, label="UInt16"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["AddressString", "Strict", "Address", "Port"]),
        # 
        'RtlIpv6AddressToStringA': SimTypeFunction([SimTypePointer(SimStruct({"u": SimUnion({"Byte": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 16), "Word": SimTypeFixedSizeArray(SimTypeShort(signed=False, label="UInt16"), 8)}, name="<anon>", label="None")}, name="IN6_ADDR", pack=False, align=None), offset=0), SimTypePointer(SimTypeChar(label="Byte"), label="LPArray", offset=0)], SimTypePointer(SimTypeChar(label="Byte"), offset=0), arg_names=["Addr", "S"]),
        # 
        'RtlIpv6AddressToStringExA': SimTypeFunction([SimTypePointer(SimStruct({"u": SimUnion({"Byte": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 16), "Word": SimTypeFixedSizeArray(SimTypeShort(signed=False, label="UInt16"), 8)}, name="<anon>", label="None")}, name="IN6_ADDR", pack=False, align=None), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypeShort(signed=False, label="UInt16"), SimTypePointer(SimTypeChar(label="Byte"), label="LPArray", offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["Address", "ScopeId", "Port", "AddressString", "AddressStringLength"]),
        # 
        'RtlIpv6AddressToStringW': SimTypeFunction([SimTypePointer(SimStruct({"u": SimUnion({"Byte": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 16), "Word": SimTypeFixedSizeArray(SimTypeShort(signed=False, label="UInt16"), 8)}, name="<anon>", label="None")}, name="IN6_ADDR", pack=False, align=None), offset=0), SimTypePointer(SimTypeChar(label="Char"), label="LPArray", offset=0)], SimTypePointer(SimTypeChar(label="Char"), offset=0), arg_names=["Addr", "S"]),
        # 
        'RtlIpv6AddressToStringExW': SimTypeFunction([SimTypePointer(SimStruct({"u": SimUnion({"Byte": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 16), "Word": SimTypeFixedSizeArray(SimTypeShort(signed=False, label="UInt16"), 8)}, name="<anon>", label="None")}, name="IN6_ADDR", pack=False, align=None), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypeShort(signed=False, label="UInt16"), SimTypePointer(SimTypeChar(label="Char"), label="LPArray", offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["Address", "ScopeId", "Port", "AddressString", "AddressStringLength"]),
        # 
        'RtlIpv6StringToAddressA': SimTypeFunction([SimTypePointer(SimTypeChar(label="Byte"), offset=0), SimTypePointer(SimTypePointer(SimTypeChar(label="Byte"), offset=0), offset=0), SimTypePointer(SimStruct({"u": SimUnion({"Byte": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 16), "Word": SimTypeFixedSizeArray(SimTypeShort(signed=False, label="UInt16"), 8)}, name="<anon>", label="None")}, name="IN6_ADDR", pack=False, align=None), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["S", "Terminator", "Addr"]),
        # 
        'RtlIpv6StringToAddressExA': SimTypeFunction([SimTypePointer(SimTypeChar(label="Byte"), offset=0), SimTypePointer(SimStruct({"u": SimUnion({"Byte": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 16), "Word": SimTypeFixedSizeArray(SimTypeShort(signed=False, label="UInt16"), 8)}, name="<anon>", label="None")}, name="IN6_ADDR", pack=False, align=None), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0), SimTypePointer(SimTypeShort(signed=False, label="UInt16"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["AddressString", "Address", "ScopeId", "Port"]),
        # 
        'RtlIpv6StringToAddressW': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypePointer(SimTypeChar(label="Char"), offset=0), offset=0), SimTypePointer(SimStruct({"u": SimUnion({"Byte": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 16), "Word": SimTypeFixedSizeArray(SimTypeShort(signed=False, label="UInt16"), 8)}, name="<anon>", label="None")}, name="IN6_ADDR", pack=False, align=None), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["S", "Terminator", "Addr"]),
        # 
        'RtlIpv6StringToAddressExW': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimStruct({"u": SimUnion({"Byte": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 16), "Word": SimTypeFixedSizeArray(SimTypeShort(signed=False, label="UInt16"), 8)}, name="<anon>", label="None")}, name="IN6_ADDR", pack=False, align=None), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0), SimTypePointer(SimTypeShort(signed=False, label="UInt16"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["AddressString", "Address", "ScopeId", "Port"]),
        # 
        'RtlEthernetAddressToStringA': SimTypeFunction([SimTypePointer(SimUnion({"Byte": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 6), "Anonymous": SimStruct({"Oui": SimUnion({"Byte": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 3), "Anonymous": SimStruct({"_bitfield": SimTypeChar(label="Byte")}, name="_Anonymous_e__Struct", pack=False, align=None)}, name="<anon>", label="None"), "Ei48": SimUnion({"Byte": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 3)}, name="<anon>", label="None")}, name="_Anonymous_e__Struct", pack=False, align=None)}, name="<anon>", label="None"), offset=0), SimTypePointer(SimTypeChar(label="Byte"), label="LPArray", offset=0)], SimTypePointer(SimTypeChar(label="Byte"), offset=0), arg_names=["Addr", "S"]),
        # 
        'RtlEthernetAddressToStringW': SimTypeFunction([SimTypePointer(SimUnion({"Byte": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 6), "Anonymous": SimStruct({"Oui": SimUnion({"Byte": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 3), "Anonymous": SimStruct({"_bitfield": SimTypeChar(label="Byte")}, name="_Anonymous_e__Struct", pack=False, align=None)}, name="<anon>", label="None"), "Ei48": SimUnion({"Byte": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 3)}, name="<anon>", label="None")}, name="_Anonymous_e__Struct", pack=False, align=None)}, name="<anon>", label="None"), offset=0), SimTypePointer(SimTypeChar(label="Char"), label="LPArray", offset=0)], SimTypePointer(SimTypeChar(label="Char"), offset=0), arg_names=["Addr", "S"]),
        # 
        'RtlEthernetStringToAddressA': SimTypeFunction([SimTypePointer(SimTypeChar(label="Byte"), offset=0), SimTypePointer(SimTypePointer(SimTypeChar(label="Byte"), offset=0), offset=0), SimTypePointer(SimUnion({"Byte": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 6), "Anonymous": SimStruct({"Oui": SimUnion({"Byte": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 3), "Anonymous": SimStruct({"_bitfield": SimTypeChar(label="Byte")}, name="_Anonymous_e__Struct", pack=False, align=None)}, name="<anon>", label="None"), "Ei48": SimUnion({"Byte": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 3)}, name="<anon>", label="None")}, name="_Anonymous_e__Struct", pack=False, align=None)}, name="<anon>", label="None"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["S", "Terminator", "Addr"]),
        # 
        'RtlEthernetStringToAddressW': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypePointer(SimTypeChar(label="Char"), offset=0), offset=0), SimTypePointer(SimUnion({"Byte": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 6), "Anonymous": SimStruct({"Oui": SimUnion({"Byte": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 3), "Anonymous": SimStruct({"_bitfield": SimTypeChar(label="Byte")}, name="_Anonymous_e__Struct", pack=False, align=None)}, name="<anon>", label="None"), "Ei48": SimUnion({"Byte": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 3)}, name="<anon>", label="None")}, name="_Anonymous_e__Struct", pack=False, align=None)}, name="<anon>", label="None"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["S", "Terminator", "Addr"]),
        # 
        'RtlConvertSidToUnicodeString': SimTypeFunction([SimTypePointer(SimStruct({"Length": SimTypeShort(signed=False, label="UInt16"), "MaximumLength": SimTypeShort(signed=False, label="UInt16"), "Buffer": SimTypePointer(SimTypeChar(label="Char"), offset=0)}, name="UNICODE_STRING", pack=False, align=None), offset=0), SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypeChar(label="Byte")], SimTypeInt(signed=True, label="Int32"), arg_names=["UnicodeString", "Sid", "AllocateDestinationString"]),
        # 
        'RtlAddGrowableFunctionTable': SimTypeFunction([SimTypePointer(SimTypePointer(SimTypeBottom(label="Void"), offset=0), offset=0), SimTypePointer(SimStruct({"BeginAddress": SimTypeInt(signed=False, label="UInt32"), "Anonymous": SimUnion({"UnwindData": SimTypeInt(signed=False, label="UInt32"), "Anonymous": SimStruct({"_bitfield": SimTypeInt(signed=False, label="UInt32")}, name="_Anonymous_e__Struct", pack=False, align=None)}, name="<anon>", label="None")}, name="IMAGE_ARM64_RUNTIME_FUNCTION_ENTRY", pack=False, align=None), label="LPArray", offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeInt(signed=False, label="UInt"), label="UIntPtr", offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt"), label="UIntPtr", offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["DynamicTable", "FunctionTable", "EntryCount", "MaximumEntryCount", "RangeBase", "RangeEnd"]),
        # 
        'RtlGrowFunctionTable': SimTypeFunction([SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypeInt(signed=False, label="UInt32")], SimTypeBottom(label="Void"), arg_names=["DynamicTable", "NewEntryCount"]),
        # 
        'RtlDeleteGrowableFunctionTable': SimTypeFunction([SimTypePointer(SimTypeBottom(label="Void"), offset=0)], SimTypeBottom(label="Void"), arg_names=["DynamicTable"]),
        # 
        'RtlAddGrowableFunctionTable': SimTypeFunction([SimTypePointer(SimTypePointer(SimTypeBottom(label="Void"), offset=0), offset=0), SimTypePointer(SimStruct({"BeginAddress": SimTypeInt(signed=False, label="UInt32"), "EndAddress": SimTypeInt(signed=False, label="UInt32"), "Anonymous": SimUnion({"UnwindInfoAddress": SimTypeInt(signed=False, label="UInt32"), "UnwindData": SimTypeInt(signed=False, label="UInt32")}, name="<anon>", label="None")}, name="IMAGE_RUNTIME_FUNCTION_ENTRY", pack=False, align=None), label="LPArray", offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeInt(signed=False, label="UInt"), label="UIntPtr", offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt"), label="UIntPtr", offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["DynamicTable", "FunctionTable", "EntryCount", "MaximumEntryCount", "RangeBase", "RangeEnd"]),
        # 
        'RtlGetNonVolatileToken': SimTypeFunction([SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt"), label="UIntPtr", offset=0), SimTypePointer(SimTypePointer(SimTypeBottom(label="Void"), offset=0), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["NvBuffer", "Size", "NvToken"]),
        # 
        'RtlFreeNonVolatileToken': SimTypeFunction([SimTypePointer(SimTypeBottom(label="Void"), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["NvToken"]),
        # 
        'RtlFlushNonVolatileMemory': SimTypeFunction([SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt"), label="UIntPtr", offset=0), SimTypeInt(signed=False, label="UInt32")], SimTypeInt(signed=False, label="UInt32"), arg_names=["NvToken", "NvBuffer", "Size", "Flags"]),
        # 
        'RtlDrainNonVolatileFlush': SimTypeFunction([SimTypePointer(SimTypeBottom(label="Void"), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["NvToken"]),
        # 
        'RtlWriteNonVolatileMemory': SimTypeFunction([SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt"), label="UIntPtr", offset=0), SimTypeInt(signed=False, label="UInt32")], SimTypeInt(signed=False, label="UInt32"), arg_names=["NvToken", "NvDestination", "Source", "Size", "Flags"]),
        # 
        'RtlFillNonVolatileMemory': SimTypeFunction([SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt"), label="UIntPtr", offset=0), SimTypeChar(label="Byte"), SimTypeInt(signed=False, label="UInt32")], SimTypeInt(signed=False, label="UInt32"), arg_names=["NvToken", "NvDestination", "Size", "Value", "Flags"]),
        # 
        'RtlFlushNonVolatileMemoryRanges': SimTypeFunction([SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypePointer(SimStruct({"BaseAddress": SimTypePointer(SimTypeBottom(label="Void"), offset=0), "Length": SimTypePointer(SimTypeInt(signed=False, label="UInt"), label="UIntPtr", offset=0)}, name="NV_MEMORY_RANGE", pack=False, align=None), label="LPArray", offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt"), label="UIntPtr", offset=0), SimTypeInt(signed=False, label="UInt32")], SimTypeInt(signed=False, label="UInt32"), arg_names=["NvToken", "NvRanges", "NumRanges", "Flags"]),
        # 
        'RtlInitializeSListHead': SimTypeFunction([SimTypePointer(SimUnion({"Alignment": SimTypeLongLong(signed=False, label="UInt64"), "Anonymous": SimStruct({"Next": SimStruct({"Next": SimTypePointer(SimTypeBottom(label="SLIST_ENTRY"), offset=0)}, name="SLIST_ENTRY", pack=False, align=None), "Depth": SimTypeShort(signed=False, label="UInt16"), "CpuId": SimTypeShort(signed=False, label="UInt16")}, name="_Anonymous_e__Struct", pack=False, align=None)}, name="<anon>", label="None"), offset=0)], SimTypeBottom(label="Void"), arg_names=["ListHead"]),
        # 
        'RtlFirstEntrySList': SimTypeFunction([SimTypePointer(SimUnion({"Alignment": SimTypeLongLong(signed=False, label="UInt64"), "Anonymous": SimStruct({"Next": SimStruct({"Next": SimTypePointer(SimTypeBottom(label="SLIST_ENTRY"), offset=0)}, name="SLIST_ENTRY", pack=False, align=None), "Depth": SimTypeShort(signed=False, label="UInt16"), "CpuId": SimTypeShort(signed=False, label="UInt16")}, name="_Anonymous_e__Struct", pack=False, align=None)}, name="<anon>", label="None"), offset=0)], SimTypePointer(SimStruct({"Next": SimTypePointer(SimTypeBottom(label="SLIST_ENTRY"), offset=0)}, name="SLIST_ENTRY", pack=False, align=None), offset=0), arg_names=["ListHead"]),
        # 
        'RtlInterlockedPopEntrySList': SimTypeFunction([SimTypePointer(SimUnion({"Alignment": SimTypeLongLong(signed=False, label="UInt64"), "Anonymous": SimStruct({"Next": SimStruct({"Next": SimTypePointer(SimTypeBottom(label="SLIST_ENTRY"), offset=0)}, name="SLIST_ENTRY", pack=False, align=None), "Depth": SimTypeShort(signed=False, label="UInt16"), "CpuId": SimTypeShort(signed=False, label="UInt16")}, name="_Anonymous_e__Struct", pack=False, align=None)}, name="<anon>", label="None"), offset=0)], SimTypePointer(SimStruct({"Next": SimTypePointer(SimTypeBottom(label="SLIST_ENTRY"), offset=0)}, name="SLIST_ENTRY", pack=False, align=None), offset=0), arg_names=["ListHead"]),
        # 
        'RtlInterlockedPushEntrySList': SimTypeFunction([SimTypePointer(SimUnion({"Alignment": SimTypeLongLong(signed=False, label="UInt64"), "Anonymous": SimStruct({"Next": SimStruct({"Next": SimTypePointer(SimTypeBottom(label="SLIST_ENTRY"), offset=0)}, name="SLIST_ENTRY", pack=False, align=None), "Depth": SimTypeShort(signed=False, label="UInt16"), "CpuId": SimTypeShort(signed=False, label="UInt16")}, name="_Anonymous_e__Struct", pack=False, align=None)}, name="<anon>", label="None"), offset=0), SimTypePointer(SimStruct({"Next": SimTypePointer(SimTypeBottom(label="SLIST_ENTRY"), offset=0)}, name="SLIST_ENTRY", pack=False, align=None), offset=0)], SimTypePointer(SimStruct({"Next": SimTypePointer(SimTypeBottom(label="SLIST_ENTRY"), offset=0)}, name="SLIST_ENTRY", pack=False, align=None), offset=0), arg_names=["ListHead", "ListEntry"]),
        # 
        'RtlInterlockedPushListSListEx': SimTypeFunction([SimTypePointer(SimUnion({"Alignment": SimTypeLongLong(signed=False, label="UInt64"), "Anonymous": SimStruct({"Next": SimStruct({"Next": SimTypePointer(SimTypeBottom(label="SLIST_ENTRY"), offset=0)}, name="SLIST_ENTRY", pack=False, align=None), "Depth": SimTypeShort(signed=False, label="UInt16"), "CpuId": SimTypeShort(signed=False, label="UInt16")}, name="_Anonymous_e__Struct", pack=False, align=None)}, name="<anon>", label="None"), offset=0), SimTypePointer(SimStruct({"Next": SimTypePointer(SimTypeBottom(label="SLIST_ENTRY"), offset=0)}, name="SLIST_ENTRY", pack=False, align=None), offset=0), SimTypePointer(SimStruct({"Next": SimTypePointer(SimTypeBottom(label="SLIST_ENTRY"), offset=0)}, name="SLIST_ENTRY", pack=False, align=None), offset=0), SimTypeInt(signed=False, label="UInt32")], SimTypePointer(SimStruct({"Next": SimTypePointer(SimTypeBottom(label="SLIST_ENTRY"), offset=0)}, name="SLIST_ENTRY", pack=False, align=None), offset=0), arg_names=["ListHead", "List", "ListEnd", "Count"]),
        # 
        'RtlInterlockedFlushSList': SimTypeFunction([SimTypePointer(SimUnion({"Alignment": SimTypeLongLong(signed=False, label="UInt64"), "Anonymous": SimStruct({"Next": SimStruct({"Next": SimTypePointer(SimTypeBottom(label="SLIST_ENTRY"), offset=0)}, name="SLIST_ENTRY", pack=False, align=None), "Depth": SimTypeShort(signed=False, label="UInt16"), "CpuId": SimTypeShort(signed=False, label="UInt16")}, name="_Anonymous_e__Struct", pack=False, align=None)}, name="<anon>", label="None"), offset=0)], SimTypePointer(SimStruct({"Next": SimTypePointer(SimTypeBottom(label="SLIST_ENTRY"), offset=0)}, name="SLIST_ENTRY", pack=False, align=None), offset=0), arg_names=["ListHead"]),
        # 
        'RtlQueryDepthSList': SimTypeFunction([SimTypePointer(SimUnion({"Alignment": SimTypeLongLong(signed=False, label="UInt64"), "Anonymous": SimStruct({"Next": SimStruct({"Next": SimTypePointer(SimTypeBottom(label="SLIST_ENTRY"), offset=0)}, name="SLIST_ENTRY", pack=False, align=None), "Depth": SimTypeShort(signed=False, label="UInt16"), "CpuId": SimTypeShort(signed=False, label="UInt16")}, name="_Anonymous_e__Struct", pack=False, align=None)}, name="<anon>", label="None"), offset=0)], SimTypeShort(signed=False, label="UInt16"), arg_names=["ListHead"]),
        # 
        'RtlGetReturnAddressHijackTarget': SimTypeFunction([], SimTypePointer(SimTypeInt(signed=False, label="UInt"), label="UIntPtr", offset=0)),
        # 
        'RtlGetProductInfo': SimTypeFunction([SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0)], SimTypeChar(label="Byte"), arg_names=["OSMajorVersion", "OSMinorVersion", "SpMajorVersion", "SpMinorVersion", "ReturnedProductType"]),
        # 
        'RtlCrc32': SimTypeFunction([SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt"), label="UIntPtr", offset=0), SimTypeInt(signed=False, label="UInt32")], SimTypeInt(signed=False, label="UInt32"), arg_names=["Buffer", "Size", "InitialCrc"]),
        # 
        'RtlCrc64': SimTypeFunction([SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt"), label="UIntPtr", offset=0), SimTypeLongLong(signed=False, label="UInt64")], SimTypeLongLong(signed=False, label="UInt64"), arg_names=["Buffer", "Size", "InitialCrc"]),
        # 
        'RtlOsDeploymentState': SimTypeFunction([SimTypeInt(signed=False, label="UInt32")], SimTypeInt(signed=False, label="OS_DEPLOYEMENT_STATE_VALUES"), arg_names=["Flags"]),
        # 
        'RtlInitializeCorrelationVector': SimTypeFunction([SimTypePointer(SimStruct({"Version": SimTypeChar(label="Byte"), "Vector": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 129)}, name="CORRELATION_VECTOR", pack=False, align=None), offset=0), SimTypeInt(signed=True, label="Int32"), SimTypePointer(SimTypeBottom(label="Guid"), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["CorrelationVector", "Version", "Guid"]),
        # 
        'RtlIncrementCorrelationVector': SimTypeFunction([SimTypePointer(SimStruct({"Version": SimTypeChar(label="Byte"), "Vector": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 129)}, name="CORRELATION_VECTOR", pack=False, align=None), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["CorrelationVector"]),
        # 
        'RtlExtendCorrelationVector': SimTypeFunction([SimTypePointer(SimStruct({"Version": SimTypeChar(label="Byte"), "Vector": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 129)}, name="CORRELATION_VECTOR", pack=False, align=None), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["CorrelationVector"]),
        # 
        'RtlValidateCorrelationVector': SimTypeFunction([SimTypePointer(SimStruct({"Version": SimTypeChar(label="Byte"), "Vector": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 129)}, name="CORRELATION_VECTOR", pack=False, align=None), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["Vector"]),
        # 
        'RtlRaiseCustomSystemEventTrigger': SimTypeFunction([SimTypePointer(SimStruct({"Size": SimTypeInt(signed=False, label="UInt32"), "TriggerId": SimTypePointer(SimTypeChar(label="Char"), offset=0)}, name="CUSTOM_SYSTEM_EVENT_TRIGGER_CONFIG", pack=False, align=None), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["TriggerConfig"]),
        # 
        'RtlIsZeroMemory': SimTypeFunction([SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt"), label="UIntPtr", offset=0)], SimTypeChar(label="Byte"), arg_names=["Buffer", "Length"]),
        # 
        'RtlNormalizeSecurityDescriptor': SimTypeFunction([SimTypePointer(SimTypePointer(SimStruct({"Revision": SimTypeChar(label="Byte"), "Sbz1": SimTypeChar(label="Byte"), "Control": SimTypeShort(signed=False, label="UInt16"), "Owner": SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), "Group": SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), "Sacl": SimTypePointer(SimStruct({"AclRevision": SimTypeChar(label="Byte"), "Sbz1": SimTypeChar(label="Byte"), "AclSize": SimTypeShort(signed=False, label="UInt16"), "AceCount": SimTypeShort(signed=False, label="UInt16"), "Sbz2": SimTypeShort(signed=False, label="UInt16")}, name="ACL", pack=False, align=None), offset=0), "Dacl": SimTypePointer(SimStruct({"AclRevision": SimTypeChar(label="Byte"), "Sbz1": SimTypeChar(label="Byte"), "AclSize": SimTypeShort(signed=False, label="UInt16"), "AceCount": SimTypeShort(signed=False, label="UInt16"), "Sbz2": SimTypeShort(signed=False, label="UInt16")}, name="ACL", pack=False, align=None), offset=0)}, name="SECURITY_DESCRIPTOR", pack=False, align=None), offset=0), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypePointer(SimStruct({"Revision": SimTypeChar(label="Byte"), "Sbz1": SimTypeChar(label="Byte"), "Control": SimTypeShort(signed=False, label="UInt16"), "Owner": SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), "Group": SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), "Sacl": SimTypePointer(SimStruct({"AclRevision": SimTypeChar(label="Byte"), "Sbz1": SimTypeChar(label="Byte"), "AclSize": SimTypeShort(signed=False, label="UInt16"), "AceCount": SimTypeShort(signed=False, label="UInt16"), "Sbz2": SimTypeShort(signed=False, label="UInt16")}, name="ACL", pack=False, align=None), offset=0), "Dacl": SimTypePointer(SimStruct({"AclRevision": SimTypeChar(label="Byte"), "Sbz1": SimTypeChar(label="Byte"), "AclSize": SimTypeShort(signed=False, label="UInt16"), "AceCount": SimTypeShort(signed=False, label="UInt16"), "Sbz2": SimTypeShort(signed=False, label="UInt16")}, name="ACL", pack=False, align=None), offset=0)}, name="SECURITY_DESCRIPTOR", pack=False, align=None), offset=0), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0), SimTypeChar(label="Byte")], SimTypeChar(label="Byte"), arg_names=["SecurityDescriptor", "SecurityDescriptorLength", "NewSecurityDescriptor", "NewSecurityDescriptorLength", "CheckOnly"]),
        # 
        'RtlGetDeviceFamilyInfoEnum': SimTypeFunction([SimTypePointer(SimTypeLongLong(signed=False, label="UInt64"), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0)], SimTypeBottom(label="Void"), arg_names=["pullUAPInfo", "pulDeviceFamily", "pulDeviceForm"]),
        # 
        'RtlConvertDeviceFamilyInfoToString': SimTypeFunction([SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeChar(label="Char"), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["pulDeviceFamilyBufferSize", "pulDeviceFormBufferSize", "DeviceFamily", "DeviceForm"]),
        # 
        'RtlSwitchedVVI': SimTypeFunction([SimTypePointer(SimStruct({"dwOSVersionInfoSize": SimTypeInt(signed=False, label="UInt32"), "dwMajorVersion": SimTypeInt(signed=False, label="UInt32"), "dwMinorVersion": SimTypeInt(signed=False, label="UInt32"), "dwBuildNumber": SimTypeInt(signed=False, label="UInt32"), "dwPlatformId": SimTypeInt(signed=False, label="UInt32"), "szCSDVersion": SimTypeFixedSizeArray(SimTypeChar(label="Char"), 128), "wServicePackMajor": SimTypeShort(signed=False, label="UInt16"), "wServicePackMinor": SimTypeShort(signed=False, label="UInt16"), "wSuiteMask": SimTypeShort(signed=False, label="UInt16"), "wProductType": SimTypeChar(label="Byte"), "wReserved": SimTypeChar(label="Byte")}, name="OSVERSIONINFOEXW", pack=False, align=None), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypeLongLong(signed=False, label="UInt64")], SimTypeInt(signed=False, label="UInt32"), arg_names=["VersionInfo", "TypeMask", "ConditionMask"]),
        # 
        'NtQueryInformationProcess': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypeInt(signed=False, label="PROCESSINFOCLASS"), SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["ProcessHandle", "ProcessInformationClass", "ProcessInformation", "ProcessInformationLength", "ReturnLength"]),
        # 
        'NtQueryInformationThread': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypeInt(signed=False, label="THREADINFOCLASS"), SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["ThreadHandle", "ThreadInformationClass", "ThreadInformation", "ThreadInformationLength", "ReturnLength"]),
        # 
        'NtClose': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["Handle"]),
        # 
        'NtCreateFile': SimTypeFunction([SimTypePointer(SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimStruct({"Length": SimTypeInt(signed=False, label="UInt32"), "RootDirectory": SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), "ObjectName": SimTypePointer(SimStruct({"Length": SimTypeShort(signed=False, label="UInt16"), "MaximumLength": SimTypeShort(signed=False, label="UInt16"), "Buffer": SimTypePointer(SimTypeChar(label="Char"), offset=0)}, name="UNICODE_STRING", pack=False, align=None), offset=0), "Attributes": SimTypeInt(signed=False, label="UInt32"), "SecurityDescriptor": SimTypePointer(SimTypeBottom(label="Void"), offset=0), "SecurityQualityOfService": SimTypePointer(SimTypeBottom(label="Void"), offset=0)}, name="OBJECT_ATTRIBUTES", pack=False, align=None), offset=0), SimTypePointer(SimStruct({"Anonymous": SimUnion({"Status": SimTypeInt(signed=True, label="Int32"), "Pointer": SimTypePointer(SimTypeBottom(label="Void"), offset=0)}, name="<anon>", label="None"), "Information": SimTypePointer(SimTypeInt(signed=False, label="UInt"), label="UIntPtr", offset=0)}, name="IO_STATUS_BLOCK", pack=False, align=None), offset=0), SimTypePointer(SimUnion({"Anonymous": SimStruct({"LowPart": SimTypeInt(signed=False, label="UInt32"), "HighPart": SimTypeInt(signed=True, label="Int32")}, name="_Anonymous_e__Struct", pack=False, align=None), "u": SimStruct({"LowPart": SimTypeInt(signed=False, label="UInt32"), "HighPart": SimTypeInt(signed=True, label="Int32")}, name="_u_e__Struct", pack=False, align=None), "QuadPart": SimTypeLongLong(signed=True, label="Int64")}, name="<anon>", label="None"), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="FILE_SHARE_MODE"), SimTypeInt(signed=False, label="NT_CREATE_FILE_DISPOSITION"), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypeInt(signed=False, label="UInt32")], SimTypeInt(signed=True, label="Int32"), arg_names=["FileHandle", "DesiredAccess", "ObjectAttributes", "IoStatusBlock", "AllocationSize", "FileAttributes", "ShareAccess", "CreateDisposition", "CreateOptions", "EaBuffer", "EaLength"]),
        # 
        'NtOpenFile': SimTypeFunction([SimTypePointer(SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimStruct({"Length": SimTypeInt(signed=False, label="UInt32"), "RootDirectory": SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), "ObjectName": SimTypePointer(SimStruct({"Length": SimTypeShort(signed=False, label="UInt16"), "MaximumLength": SimTypeShort(signed=False, label="UInt16"), "Buffer": SimTypePointer(SimTypeChar(label="Char"), offset=0)}, name="UNICODE_STRING", pack=False, align=None), offset=0), "Attributes": SimTypeInt(signed=False, label="UInt32"), "SecurityDescriptor": SimTypePointer(SimTypeBottom(label="Void"), offset=0), "SecurityQualityOfService": SimTypePointer(SimTypeBottom(label="Void"), offset=0)}, name="OBJECT_ATTRIBUTES", pack=False, align=None), offset=0), SimTypePointer(SimStruct({"Anonymous": SimUnion({"Status": SimTypeInt(signed=True, label="Int32"), "Pointer": SimTypePointer(SimTypeBottom(label="Void"), offset=0)}, name="<anon>", label="None"), "Information": SimTypePointer(SimTypeInt(signed=False, label="UInt"), label="UIntPtr", offset=0)}, name="IO_STATUS_BLOCK", pack=False, align=None), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32")], SimTypeInt(signed=True, label="Int32"), arg_names=["FileHandle", "DesiredAccess", "ObjectAttributes", "IoStatusBlock", "ShareAccess", "OpenOptions"]),
        # 
        'NtRenameKey': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypePointer(SimStruct({"Length": SimTypeShort(signed=False, label="UInt16"), "MaximumLength": SimTypeShort(signed=False, label="UInt16"), "Buffer": SimTypePointer(SimTypeChar(label="Char"), offset=0)}, name="UNICODE_STRING", pack=False, align=None), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["KeyHandle", "NewName"]),
        # 
        'NtNotifyChangeMultipleKeys': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimStruct({"Length": SimTypeInt(signed=False, label="UInt32"), "RootDirectory": SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), "ObjectName": SimTypePointer(SimStruct({"Length": SimTypeShort(signed=False, label="UInt16"), "MaximumLength": SimTypeShort(signed=False, label="UInt16"), "Buffer": SimTypePointer(SimTypeChar(label="Char"), offset=0)}, name="UNICODE_STRING", pack=False, align=None), offset=0), "Attributes": SimTypeInt(signed=False, label="UInt32"), "SecurityDescriptor": SimTypePointer(SimTypeBottom(label="Void"), offset=0), "SecurityQualityOfService": SimTypePointer(SimTypeBottom(label="Void"), offset=0)}, name="OBJECT_ATTRIBUTES", pack=False, align=None), label="LPArray", offset=0), SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypePointer(SimTypeFunction([SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypePointer(SimStruct({"Anonymous": SimUnion({"Status": SimTypeInt(signed=True, label="Int32"), "Pointer": SimTypePointer(SimTypeBottom(label="Void"), offset=0)}, name="<anon>", label="None"), "Information": SimTypePointer(SimTypeInt(signed=False, label="UInt"), label="UIntPtr", offset=0)}, name="IO_STATUS_BLOCK", pack=False, align=None), offset=0), SimTypeInt(signed=False, label="UInt32")], SimTypeBottom(label="Void"), arg_names=["ApcContext", "IoStatusBlock", "Reserved"]), offset=0), SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypePointer(SimStruct({"Anonymous": SimUnion({"Status": SimTypeInt(signed=True, label="Int32"), "Pointer": SimTypePointer(SimTypeBottom(label="Void"), offset=0)}, name="<anon>", label="None"), "Information": SimTypePointer(SimTypeInt(signed=False, label="UInt"), label="UIntPtr", offset=0)}, name="IO_STATUS_BLOCK", pack=False, align=None), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypeChar(label="Byte"), SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypeChar(label="Byte")], SimTypeInt(signed=True, label="Int32"), arg_names=["MasterKeyHandle", "Count", "SubordinateObjects", "Event", "ApcRoutine", "ApcContext", "IoStatusBlock", "CompletionFilter", "WatchTree", "Buffer", "BufferSize", "Asynchronous"]),
        # 
        'NtQueryMultipleValueKey': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypePointer(SimStruct({"ValueName": SimTypePointer(SimStruct({"Length": SimTypeShort(signed=False, label="UInt16"), "MaximumLength": SimTypeShort(signed=False, label="UInt16"), "Buffer": SimTypePointer(SimTypeChar(label="Char"), offset=0)}, name="UNICODE_STRING", pack=False, align=None), offset=0), "DataLength": SimTypeInt(signed=False, label="UInt32"), "DataOffset": SimTypeInt(signed=False, label="UInt32"), "Type": SimTypeInt(signed=False, label="UInt32")}, name="KEY_VALUE_ENTRY", pack=False, align=None), label="LPArray", offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["KeyHandle", "ValueEntries", "EntryCount", "ValueBuffer", "BufferLength", "RequiredBufferLength"]),
        # 
        'NtSetInformationKey': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypeInt(signed=False, label="KEY_SET_INFORMATION_CLASS"), SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypeInt(signed=False, label="UInt32")], SimTypeInt(signed=True, label="Int32"), arg_names=["KeyHandle", "KeySetInformationClass", "KeySetInformation", "KeySetInformationLength"]),
        # 
        'NtDeviceIoControlFile': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypePointer(SimTypeFunction([SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypePointer(SimStruct({"Anonymous": SimUnion({"Status": SimTypeInt(signed=True, label="Int32"), "Pointer": SimTypePointer(SimTypeBottom(label="Void"), offset=0)}, name="<anon>", label="None"), "Information": SimTypePointer(SimTypeInt(signed=False, label="UInt"), label="UIntPtr", offset=0)}, name="IO_STATUS_BLOCK", pack=False, align=None), offset=0), SimTypeInt(signed=False, label="UInt32")], SimTypeBottom(label="Void"), arg_names=["ApcContext", "IoStatusBlock", "Reserved"]), offset=0), SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypePointer(SimStruct({"Anonymous": SimUnion({"Status": SimTypeInt(signed=True, label="Int32"), "Pointer": SimTypePointer(SimTypeBottom(label="Void"), offset=0)}, name="<anon>", label="None"), "Information": SimTypePointer(SimTypeInt(signed=False, label="UInt"), label="UIntPtr", offset=0)}, name="IO_STATUS_BLOCK", pack=False, align=None), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypeInt(signed=False, label="UInt32")], SimTypeInt(signed=True, label="Int32"), arg_names=["FileHandle", "Event", "ApcRoutine", "ApcContext", "IoStatusBlock", "IoControlCode", "InputBuffer", "InputBufferLength", "OutputBuffer", "OutputBufferLength"]),
        # 
        'NtWaitForSingleObject': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypeChar(label="Byte"), SimTypePointer(SimUnion({"Anonymous": SimStruct({"LowPart": SimTypeInt(signed=False, label="UInt32"), "HighPart": SimTypeInt(signed=True, label="Int32")}, name="_Anonymous_e__Struct", pack=False, align=None), "u": SimStruct({"LowPart": SimTypeInt(signed=False, label="UInt32"), "HighPart": SimTypeInt(signed=True, label="Int32")}, name="_u_e__Struct", pack=False, align=None), "QuadPart": SimTypeLongLong(signed=True, label="Int64")}, name="<anon>", label="None"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["Handle", "Alertable", "Timeout"]),
        # 
        'RtlIsNameLegalDOS8Dot3': SimTypeFunction([SimTypePointer(SimStruct({"Length": SimTypeShort(signed=False, label="UInt16"), "MaximumLength": SimTypeShort(signed=False, label="UInt16"), "Buffer": SimTypePointer(SimTypeChar(label="Char"), offset=0)}, name="UNICODE_STRING", pack=False, align=None), offset=0), SimTypePointer(SimStruct({"Length": SimTypeShort(signed=False, label="UInt16"), "MaximumLength": SimTypeShort(signed=False, label="UInt16"), "Buffer": SimTypePointer(SimTypeChar(label="Byte"), offset=0)}, name="STRING", pack=False, align=None), offset=0), SimTypePointer(SimTypeChar(label="Byte"), offset=0)], SimTypeChar(label="Byte"), arg_names=["Name", "OemName", "NameContainsSpaces"]),
        # 
        'NtQueryObject': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypeInt(signed=False, label="OBJECT_INFORMATION_CLASS"), SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["Handle", "ObjectInformationClass", "ObjectInformation", "ObjectInformationLength", "ReturnLength"]),
        # 
        'NtQuerySystemInformation': SimTypeFunction([SimTypeInt(signed=False, label="SYSTEM_INFORMATION_CLASS"), SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["SystemInformationClass", "SystemInformation", "SystemInformationLength", "ReturnLength"]),
        # 
        'NtQuerySystemTime': SimTypeFunction([SimTypePointer(SimUnion({"Anonymous": SimStruct({"LowPart": SimTypeInt(signed=False, label="UInt32"), "HighPart": SimTypeInt(signed=True, label="Int32")}, name="_Anonymous_e__Struct", pack=False, align=None), "u": SimStruct({"LowPart": SimTypeInt(signed=False, label="UInt32"), "HighPart": SimTypeInt(signed=True, label="Int32")}, name="_u_e__Struct", pack=False, align=None), "QuadPart": SimTypeLongLong(signed=True, label="Int64")}, name="<anon>", label="None"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["SystemTime"]),
        # 
        'RtlLocalTimeToSystemTime': SimTypeFunction([SimTypePointer(SimUnion({"Anonymous": SimStruct({"LowPart": SimTypeInt(signed=False, label="UInt32"), "HighPart": SimTypeInt(signed=True, label="Int32")}, name="_Anonymous_e__Struct", pack=False, align=None), "u": SimStruct({"LowPart": SimTypeInt(signed=False, label="UInt32"), "HighPart": SimTypeInt(signed=True, label="Int32")}, name="_u_e__Struct", pack=False, align=None), "QuadPart": SimTypeLongLong(signed=True, label="Int64")}, name="<anon>", label="None"), offset=0), SimTypePointer(SimUnion({"Anonymous": SimStruct({"LowPart": SimTypeInt(signed=False, label="UInt32"), "HighPart": SimTypeInt(signed=True, label="Int32")}, name="_Anonymous_e__Struct", pack=False, align=None), "u": SimStruct({"LowPart": SimTypeInt(signed=False, label="UInt32"), "HighPart": SimTypeInt(signed=True, label="Int32")}, name="_u_e__Struct", pack=False, align=None), "QuadPart": SimTypeLongLong(signed=True, label="Int64")}, name="<anon>", label="None"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["LocalTime", "SystemTime"]),
        # 
        'RtlTimeToSecondsSince1970': SimTypeFunction([SimTypePointer(SimUnion({"Anonymous": SimStruct({"LowPart": SimTypeInt(signed=False, label="UInt32"), "HighPart": SimTypeInt(signed=True, label="Int32")}, name="_Anonymous_e__Struct", pack=False, align=None), "u": SimStruct({"LowPart": SimTypeInt(signed=False, label="UInt32"), "HighPart": SimTypeInt(signed=True, label="Int32")}, name="_u_e__Struct", pack=False, align=None), "QuadPart": SimTypeLongLong(signed=True, label="Int64")}, name="<anon>", label="None"), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0)], SimTypeChar(label="Byte"), arg_names=["Time", "ElapsedSeconds"]),
        # 
        'RtlFreeAnsiString': SimTypeFunction([SimTypePointer(SimStruct({"Length": SimTypeShort(signed=False, label="UInt16"), "MaximumLength": SimTypeShort(signed=False, label="UInt16"), "Buffer": SimTypePointer(SimTypeChar(label="Byte"), offset=0)}, name="STRING", pack=False, align=None), offset=0)], SimTypeBottom(label="Void"), arg_names=["AnsiString"]),
        # 
        'RtlFreeUnicodeString': SimTypeFunction([SimTypePointer(SimStruct({"Length": SimTypeShort(signed=False, label="UInt16"), "MaximumLength": SimTypeShort(signed=False, label="UInt16"), "Buffer": SimTypePointer(SimTypeChar(label="Char"), offset=0)}, name="UNICODE_STRING", pack=False, align=None), offset=0)], SimTypeBottom(label="Void"), arg_names=["UnicodeString"]),
        # 
        'RtlFreeOemString': SimTypeFunction([SimTypePointer(SimStruct({"Length": SimTypeShort(signed=False, label="UInt16"), "MaximumLength": SimTypeShort(signed=False, label="UInt16"), "Buffer": SimTypePointer(SimTypeChar(label="Byte"), offset=0)}, name="STRING", pack=False, align=None), offset=0)], SimTypeBottom(label="Void"), arg_names=["OemString"]),
        # 
        'RtlInitString': SimTypeFunction([SimTypePointer(SimStruct({"Length": SimTypeShort(signed=False, label="UInt16"), "MaximumLength": SimTypeShort(signed=False, label="UInt16"), "Buffer": SimTypePointer(SimTypeChar(label="Byte"), offset=0)}, name="STRING", pack=False, align=None), offset=0), SimTypePointer(SimTypeChar(label="SByte"), offset=0)], SimTypeBottom(label="Void"), arg_names=["DestinationString", "SourceString"]),
        # 
        'RtlInitStringEx': SimTypeFunction([SimTypePointer(SimStruct({"Length": SimTypeShort(signed=False, label="UInt16"), "MaximumLength": SimTypeShort(signed=False, label="UInt16"), "Buffer": SimTypePointer(SimTypeChar(label="Byte"), offset=0)}, name="STRING", pack=False, align=None), offset=0), SimTypePointer(SimTypeChar(label="SByte"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["DestinationString", "SourceString"]),
        # 
        'RtlInitAnsiString': SimTypeFunction([SimTypePointer(SimStruct({"Length": SimTypeShort(signed=False, label="UInt16"), "MaximumLength": SimTypeShort(signed=False, label="UInt16"), "Buffer": SimTypePointer(SimTypeChar(label="Byte"), offset=0)}, name="STRING", pack=False, align=None), offset=0), SimTypePointer(SimTypeChar(label="SByte"), offset=0)], SimTypeBottom(label="Void"), arg_names=["DestinationString", "SourceString"]),
        # 
        'RtlInitAnsiStringEx': SimTypeFunction([SimTypePointer(SimStruct({"Length": SimTypeShort(signed=False, label="UInt16"), "MaximumLength": SimTypeShort(signed=False, label="UInt16"), "Buffer": SimTypePointer(SimTypeChar(label="Byte"), offset=0)}, name="STRING", pack=False, align=None), offset=0), SimTypePointer(SimTypeChar(label="SByte"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["DestinationString", "SourceString"]),
        # 
        'RtlInitUnicodeString': SimTypeFunction([SimTypePointer(SimStruct({"Length": SimTypeShort(signed=False, label="UInt16"), "MaximumLength": SimTypeShort(signed=False, label="UInt16"), "Buffer": SimTypePointer(SimTypeChar(label="Char"), offset=0)}, name="UNICODE_STRING", pack=False, align=None), offset=0), SimTypePointer(SimTypeChar(label="Char"), offset=0)], SimTypeBottom(label="Void"), arg_names=["DestinationString", "SourceString"]),
        # 
        'RtlAnsiStringToUnicodeString': SimTypeFunction([SimTypePointer(SimStruct({"Length": SimTypeShort(signed=False, label="UInt16"), "MaximumLength": SimTypeShort(signed=False, label="UInt16"), "Buffer": SimTypePointer(SimTypeChar(label="Char"), offset=0)}, name="UNICODE_STRING", pack=False, align=None), offset=0), SimTypePointer(SimStruct({"Length": SimTypeShort(signed=False, label="UInt16"), "MaximumLength": SimTypeShort(signed=False, label="UInt16"), "Buffer": SimTypePointer(SimTypeChar(label="Byte"), offset=0)}, name="STRING", pack=False, align=None), offset=0), SimTypeChar(label="Byte")], SimTypeInt(signed=True, label="Int32"), arg_names=["DestinationString", "SourceString", "AllocateDestinationString"]),
        # 
        'RtlUnicodeStringToAnsiString': SimTypeFunction([SimTypePointer(SimStruct({"Length": SimTypeShort(signed=False, label="UInt16"), "MaximumLength": SimTypeShort(signed=False, label="UInt16"), "Buffer": SimTypePointer(SimTypeChar(label="Byte"), offset=0)}, name="STRING", pack=False, align=None), offset=0), SimTypePointer(SimStruct({"Length": SimTypeShort(signed=False, label="UInt16"), "MaximumLength": SimTypeShort(signed=False, label="UInt16"), "Buffer": SimTypePointer(SimTypeChar(label="Char"), offset=0)}, name="UNICODE_STRING", pack=False, align=None), offset=0), SimTypeChar(label="Byte")], SimTypeInt(signed=True, label="Int32"), arg_names=["DestinationString", "SourceString", "AllocateDestinationString"]),
        # 
        'RtlUnicodeStringToOemString': SimTypeFunction([SimTypePointer(SimStruct({"Length": SimTypeShort(signed=False, label="UInt16"), "MaximumLength": SimTypeShort(signed=False, label="UInt16"), "Buffer": SimTypePointer(SimTypeChar(label="Byte"), offset=0)}, name="STRING", pack=False, align=None), offset=0), SimTypePointer(SimStruct({"Length": SimTypeShort(signed=False, label="UInt16"), "MaximumLength": SimTypeShort(signed=False, label="UInt16"), "Buffer": SimTypePointer(SimTypeChar(label="Char"), offset=0)}, name="UNICODE_STRING", pack=False, align=None), offset=0), SimTypeChar(label="Byte")], SimTypeInt(signed=True, label="Int32"), arg_names=["DestinationString", "SourceString", "AllocateDestinationString"]),
        # 
        'RtlUnicodeToMultiByteSize': SimTypeFunction([SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypeInt(signed=False, label="UInt32")], SimTypeInt(signed=True, label="Int32"), arg_names=["BytesInMultiByteString", "UnicodeString", "BytesInUnicodeString"]),
        # 
        'RtlCharToInteger': SimTypeFunction([SimTypePointer(SimTypeChar(label="SByte"), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["String", "Base", "Value"]),
        # 
        'RtlUniform': SimTypeFunction([SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["Seed"]),
    }

lib.set_prototypes(prototypes)
